/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * ACPInterpolator - implementation of Adaptive Color Planes interpolation,
 * based on description given in ``Interpolation Methods for the Bayer
 * Color Array'' by Rajeev Ramanath. See copyrights.txt for patent info.
******************************************************************************/

#include "stdafx.h"

#include <algorithm>
#include <cmath>

#include "ACPInterpolator.h"
#include "Image.h"
#include "RawImageInfo.h"
#include "logstream.h"

//-----------------------------------------------------------------------------
#ifdef  _MSC_VER
#define for if(0); else for
#endif
//-----------------------------------------------------------------------------
ACPInterpolator::ACPInterpolator(): IParametrizable("ACPInterpolator")
  {
  m_small_grad_threshold = 0;
  m_small_grad_diff = 300;
  }
//-----------------------------------------------------------------------------
ACPInterpolator& ACPInterpolator::GetInterpolator()
  {
  static ACPInterpolator acp_int;
  return acp_int;
  }
//-----------------------------------------------------------------------------
void ACPInterpolator::GetParameters(Params& o_params) const
  {
  o_params["ACP_SmallGradThreshold"] = boost::lexical_cast<std::string>(m_small_grad_threshold);
  o_params["ACP_SmallGradDiff"] = boost::lexical_cast<std::string>(m_small_grad_diff);
  }
//-----------------------------------------------------------------------------
void ACPInterpolator::GetDescriptions(Params& o_params) const
  {
  o_params["ACP_SmallGradThreshold"] = "If gradients are smaller "
    "than this threshold value - undirectional interpolation will be used for current pixel";
  o_params["ACP_SmallGradDiff"] = "If gradients difference is smaller "
    "than this threshold value - undirectional interpolation will be used for current pixel";
  }
//-----------------------------------------------------------------------------
void ACPInterpolator::SetParameters(Params& i_params)
  {
  if (i_params.exist("ACP_SmallGradThreshold"))
    m_small_grad_threshold = boost::lexical_cast<ulong>(i_params.pop("ACP_SmallGradThreshold"));
  if (i_params.exist("ACP_SmallGradDiff"))
    m_small_grad_diff = boost::lexical_cast<ulong>(i_params.pop("ACP_SmallGradDiff"));
  }
//-----------------------------------------------------------------------------
void ACPInterpolator::InterpolateGreenPixel(Image& image, unsigned row, unsigned col)
  {
  uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);
  assert (ind != 1);

  int alpha_green = abs((int)image.At(row, col - 1)[1] - (int)image.At(row, col + 1)[1]);
  int alpha = alpha_green +
	      abs(-(int)image.At(row, col - 2)[ind] + 2*(int)image.At(row, col)[ind] -
		  (int)image.At(row, col + 2)[ind]);

  int beta_green = abs((int)image.At(row - 1, col)[1] - (int)image.At(row + 1, col)[1]);
  int beta  = beta_green +
	      abs(-(int)image.At(row - 2, col)[ind] + 2*(int)image.At(row, col)[ind] -
		  (int)image.At(row + 2, col)[ind]);

  int green;

  if (alpha_green < m_small_grad_threshold && beta_green < m_small_grad_threshold)
    {
    green = 
      ((int)image.At(row - 1, col)[1] + (int)image.At(row + 1, col)[1] +
       (int)image.At(row, col - 1)[1] + (int)image.At(row, col + 1)[1]) / 4;
    }
  else if (abs(alpha - beta) < m_small_grad_diff)
    green = 
      ((int)image.At(row - 1, col)[1] + (int)image.At(row + 1, col)[1] +
       (int)image.At(row, col - 1)[1] + (int)image.At(row, col + 1)[1]) / 4 +
      (-(int)image.At(row - 2, col)[ind] - (int)image.At(row + 2, col)[ind] -
	(int)image.At(row, col - 2)[ind] - (int)image.At(row, col + 2)[ind] +
	4*(int)image.At(row, col)[ind]) / 8;
  else if (alpha < beta)
    green = 
      ((int)image.At(row, col - 1)[1] + (int)image.At(row, col + 1)[1]) / 2 +
      (-(int)image.At(row, col - 2)[ind] + 2*(int)image.At(row, col)[ind] -
	(int)image.At(row, col + 2)[ind]) / 4;
  else
    green = 
      ((int)image.At(row - 1, col)[1] + (int)image.At(row + 1, col)[1]) / 2 +
      (-(int)image.At(row - 2, col)[ind] + 2*(int)image.At(row, col)[ind] -
	(int)image.At(row + 2, col)[ind]) / 4;

  image.At(row, col)[1] = image.ClampGreenPixel(green, row, col);
  }
//-----------------------------------------------------------------------------
void ACPInterpolator::InterpolatePixelInLine(Image& image, unsigned row, unsigned col)
  {
  assert (mp_image_info -> GetCurrentColorIndex(row, col) == 1);

  int red, blue;

  if (mp_image_info -> GetCurrentColorIndex(row, col - 1) == 0)
    {
    assert (mp_image_info -> GetCurrentColorIndex(row, col + 1) == 0);
    assert (mp_image_info -> GetCurrentColorIndex(row - 1, col) == 2);
    assert (mp_image_info -> GetCurrentColorIndex(row + 1, col) == 2);

    red = ((int)image.At(row, col - 1)[0] +
	   (int)image.At(row, col + 1)[0]) / 2 +
	  (-(int)image.At(row, col - 1)[1] +
	   2*(int)image.At(row, col)[1] -
	   (int)image.At(row, col + 1)[1]) / 4;

    blue = ((int)image.At(row - 1, col)[2] +
	    (int)image.At(row + 1, col)[2]) / 2 +
	   (-(int)image.At(row - 1, col)[1] +
	    2*(int)image.At(row, col)[1] -
	    (int)image.At(row + 1, col)[1]) / 4;
    }
  else
    {
    assert (mp_image_info -> GetCurrentColorIndex(row, col - 1) == 2);
    assert (mp_image_info -> GetCurrentColorIndex(row, col + 1) == 2);
    assert (mp_image_info -> GetCurrentColorIndex(row - 1, col) == 0);
    assert (mp_image_info -> GetCurrentColorIndex(row + 1, col) == 0);

    blue = ((int)image.At(row, col - 1)[2] +
	   (int)image.At(row, col + 1)[2]) / 2 +
	  (-(int)image.At(row, col - 1)[1] +
	   2*(int)image.At(row, col)[1] -
	   (int)image.At(row, col + 1)[1]) / 4;

    red = ((int)image.At(row - 1, col)[0] +
	    (int)image.At(row + 1, col)[0]) / 2 +
	   (-(int)image.At(row - 1, col)[1] +
	    2*(int)image.At(row, col)[1] -
	    (int)image.At(row + 1, col)[1]) / 4;
    }

  image.At(row, col)[0] = image.ClampChromaPixel(red, row, col);
  image.At(row, col)[2] = image.ClampChromaPixel(blue, row, col);
  }
//-----------------------------------------------------------------------------
void ACPInterpolator::InterpolatePixelInNbhd(Image& image, unsigned row, unsigned col)
  {
  uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);
  assert (ind != 1);

  uchar ind1 = (ind+2)%4;

  int alpha_chroma = abs((int)image.At(row - 1, col + 1)[ind1] -
			 (int)image.At(row + 1, col - 1)[ind1]);
  int alpha = alpha_chroma +
	      abs(-(int)image.At(row - 1, col + 1)[1] +
		  2*(int)image.At(row, col)[1] -
		  (int)image.At(row + 1, col - 1)[1]);

  int beta_chroma = abs((int)image.At(row - 1, col - 1)[ind1] -
			(int)image.At(row + 1, col + 1)[ind1]);
  int beta  = beta_chroma +
	      abs(-(int)image.At(row - 1, col - 1)[1] +
		  2*(int)image.At(row, col)[1] -
		  (int)image.At(row + 1, col + 1)[1]);

  int color;

  if (alpha < m_small_grad_threshold && beta < m_small_grad_threshold)
    color = ((int)image.At(row - 1, col - 1)[ind1] +
	     (int)image.At(row - 1, col + 1)[ind1] +
	     (int)image.At(row + 1, col - 1)[ind1] +
	     (int)image.At(row + 1, col + 1)[ind1]) / 4;
  else if (abs(alpha - beta) < m_small_grad_diff)
    color = ((int)image.At(row - 1, col - 1)[ind1] +
	     (int)image.At(row - 1, col + 1)[ind1] +
	     (int)image.At(row + 1, col - 1)[ind1] +
	     (int)image.At(row + 1, col + 1)[ind1]) / 4 +
	    (-(int)image.At(row - 1, col - 1)[1] -
	     (int)image.At(row - 1, col + 1)[1] -
	     (int)image.At(row + 1, col - 1)[1] -
	     (int)image.At(row + 1, col + 1)[1] +
	     4*(int)image.At(row, col)[1]) / 8;
  else if (alpha < beta)
    color = ((int)image.At(row - 1, col + 1)[ind1] +
	     (int)image.At(row + 1, col - 1)[ind1]) / 2 + 
	    (-(int)image.At(row - 1, col + 1)[1] -
	      (int)image.At(row + 1, col - 1)[1] +
	     2*(int)image.At(row, col)[1]) / 4;
  else
    color = ((int)image.At(row - 1, col - 1)[ind1] +
	     (int)image.At(row + 1, col + 1)[ind1]) / 2 + 
	    (-(int)image.At(row - 1, col - 1)[1] -
	      (int)image.At(row + 1, col + 1)[1] +
	     2*(int)image.At(row, col)[1]) / 4;

  image.At(row, col)[ind1] = image.ClampChromaPixel(color, row, col);
  }
//-----------------------------------------------------------------------------
bool ACPInterpolator::Interpolate(Image& image, const RawImageInfo& image_info)
  {
  mg_log << INFO << "Running ACP interpolation" << endlog;
  mp_image_info = &image_info;

  mg_log << TRACE << "Interpolating green" << endlog;

  for (unsigned row = 2; row < image.GetHeight() - 2; ++row)
    for (unsigned col = 2; col < image.GetWidth() - 2; ++col)
      {
      uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);
      if (ind != 1)
	    InterpolateGreenPixel(image, row, col);
      }

  image.SetState(Image::FULL_GREEN_COLORS);

  mg_log << TRACE << "Interpolating chroma" << endlog;

  for (unsigned row = 2; row < image.GetHeight() - 2; ++row)
    for (unsigned col = 2; col < image.GetWidth() - 2; ++col)
      {
      uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);

      if (ind == 1)
	    InterpolatePixelInLine(image, row, col);
      else
	    InterpolatePixelInNbhd(image, row, col);
      }

  image.SetState(Image::FULL_COLORS);

  mg_log << TRACE << "Finished" << endlog;

  return true;
  }
//-----------------------------------------------------------------------------
