/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * BilinearInterpolator - implementation of simple bilibnear interpolation.
******************************************************************************/

#include "stdafx.h"

#include "BilinearInterpolator.h"
#include "Image.h"
#include "RawImageInfo.h"
//-----------------------------------------------------------------------------
BilinearInterpolator::BilinearInterpolator(): IParametrizable("BilinearInterpolator")
  {
  m_fl_only_chroma = false;
  m_nbhd_size = 1;
  }
//-----------------------------------------------------------------------------
BilinearInterpolator& BilinearInterpolator::GetInterpolator()
  {
  static BilinearInterpolator b_int;
  return b_int;
  }
//-----------------------------------------------------------------------------
void BilinearInterpolator::GetParameters(Params& o_params) const
  {
  o_params["Bilinear_FlOnlyChroma"] = boost::lexical_cast<std::string>(m_fl_only_chroma);
  o_params["Bilinear_NbhdSize"] = boost::lexical_cast<std::string>(m_nbhd_size);
  }
//-----------------------------------------------------------------------------
void BilinearInterpolator::GetDescriptions(Params& o_params) const
  {
  o_params["Bilinear_FlOnlyChroma"] = "If true - interpolate only chroma channels (for internal use only)";
  o_params["Bilinear_NbhdSize"] = "Size of neighborhood to consider";
  }
//-----------------------------------------------------------------------------
void BilinearInterpolator::SetParameters(Params& i_params)
  {
  if (i_params.exist("Bilinear_FlOnlyChroma"))
    m_fl_only_chroma = boost::lexical_cast<bool>(i_params.pop("Bilinear_FlOnlyChroma"));
  if (i_params.exist("Bilinear_NbhdSize"))
    m_nbhd_size = boost::lexical_cast<ulong>(i_params.pop("Bilinear_NbhdSize"));
  }
//-----------------------------------------------------------------------------
void BilinearInterpolator::InterpolateGreenPixel(Image& image, unsigned row, unsigned col)
  {
  uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);
  assert (ind != 1);
  uchar ind1 = (ind+2)%4;

  int green, color;

  if (m_nbhd_size == 1)
    {
    green = ((int)image.At(row, col - 1)[1] +
	     (int)image.At(row, col + 1)[1] +
	     (int)image.At(row - 1, col)[1] +
	     (int)image.At(row + 1, col)[1]) / 4;

    color = ((int)image.At(row - 1, col - 1)[ind1] +
	     (int)image.At(row - 1, col + 1)[ind1] +
	     (int)image.At(row + 1, col - 1)[ind1] +
	     (int)image.At(row + 1, col + 1)[ind1]) / 4;
    }
  else
    {
    long g = 0, cc = 0;
    ulong g_count = 0, c_count = 0;

    for (unsigned r = row - m_nbhd_size; r <= row + m_nbhd_size; ++r)
      for (unsigned c = col - m_nbhd_size; c <= col + m_nbhd_size; ++c)
	{
	if (mp_image_info -> GetCurrentColorIndex(r, c) == 1)
	  {
	  g += image.At(r, c)[1];
	  ++g_count;
	  }
	else if (mp_image_info -> GetCurrentColorIndex(r, c) == ind1)
	  {
	  cc += image.At(r, c)[ind1];
	  ++c_count;
	  }
	}

    green = g/g_count;
    color = cc/c_count;
    }

  if (!m_fl_only_chroma)
    image.At(row, col)[1] = image.ClampGreenPixel(green, row, col);

  image.At(row, col)[ind1] = image.ClampChromaPixel(color, row, col);
  }
//-----------------------------------------------------------------------------
void BilinearInterpolator::InterpolateChromaPixel(Image& image, unsigned row, unsigned col)
  {
  assert (mp_image_info -> GetCurrentColorIndex(row, col) == 1);

  int red, blue;

  if (m_nbhd_size == 1)
    {
    if (mp_image_info -> GetCurrentColorIndex(row, col - 1) == 0)
      {
      red = ((int)image.At(row, col - 1)[0] +
	     (int)image.At(row, col + 1)[0]) / 2;

      blue = ((int)image.At(row - 1, col)[2] +
	     (int)image.At(row + 1, col)[2]) / 2;
      }
    else
      {
      blue = ((int)image.At(row, col - 1)[2] +
	      (int)image.At(row, col + 1)[2]) / 2;

      red = ((int)image.At(row - 1, col)[0] +
	     (int)image.At(row + 1, col)[0]) / 2;
      }
    }
  else
    {
    long rc = 0, bc = 0;
    ulong r_count = 0, b_count = 0;

    for (unsigned r = row - m_nbhd_size; r <= row + m_nbhd_size; ++r)
      for (unsigned c = col - m_nbhd_size; c <= col + m_nbhd_size; ++c)
	{
	if (mp_image_info -> GetCurrentColorIndex(r, c) == 0)
	  {
	  rc += image.At(r, c)[0];
	  ++r_count;
	  }
	else if (mp_image_info -> GetCurrentColorIndex(r, c) == 2)
	  {
	  bc += image.At(r, c)[2];
	  ++b_count;
	  }
	}

    red = rc/r_count;
    blue = bc/b_count;
    }

  image.At(row, col)[0] = image.ClampChromaPixel(red, row, col);
  image.At(row, col)[2] = image.ClampChromaPixel(blue, row, col);
  }
//-----------------------------------------------------------------------------
bool BilinearInterpolator::Interpolate(Image& image, const RawImageInfo& image_info)
  {
  mg_log << INFO << "Running Bilinear interpolation" << endlog;
  mp_image_info = &image_info;

  for (unsigned row = m_nbhd_size; row < image.GetHeight() - m_nbhd_size; ++row)
    for (unsigned col = m_nbhd_size; col < image.GetWidth() - m_nbhd_size; ++col)
      {
      uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);

      if (ind != 1)
	InterpolateGreenPixel(image, row, col);
      else
	InterpolateChromaPixel(image, row, col);
      }

  image.SetState(Image::FULL_COLORS);

  mg_log << TRACE << "Finished" << endlog;

  return true;
  }
//-----------------------------------------------------------------------------
