/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * ChromaDiffImageProcessor - implementation of noise filtering using
 * anisotropic diffusion flow performed separately for chroma components and
 * for brightness, based on description given in ``Color Image Enhancement
 * via Chromaticity Diffusion'' by Bei Tang, Guillermo Sapiro, and Vicent Caselles.
******************************************************************************/

#include "stdafx.h"

#include "ChromaDiffFlowImageProcessor.h"
#include "Image.h"
//-----------------------------------------------------------------------------
ChromaDiffFlowImageProcessor::ChromaDiffFlowImageProcessor(): IParametrizable("ChromaDiffFlowImageProcessor")
  {
  m_amount_brightness = 350;
  m_amount_chroma = 0.02;
  m_diff_num_of_iter = 10;

  /*
  m_amount_brightness = 350;
  m_amount_chroma = 0.02;
  m_diff_num_of_iter = 10;
  */
  /*
  m_amount_brightness = 2000;
  m_amount_chroma = 0.1;
  m_diff_num_of_iter = 150;
  */
  }
//-----------------------------------------------------------------------------
ChromaDiffFlowImageProcessor& ChromaDiffFlowImageProcessor::GetImageProcessor()
  {
  static ChromaDiffFlowImageProcessor cd_ip;
  return cd_ip;
  }
//-----------------------------------------------------------------------------
void ChromaDiffFlowImageProcessor::GetParameters(Params& o_params) const
  {
  o_params["ChromaFlow_AmountBrightness"] = boost::lexical_cast<std::string>(m_amount_brightness);
  o_params["ChromaFlow_DiffusionIterations"] = boost::lexical_cast<std::string>(m_diff_num_of_iter);
  }
//-----------------------------------------------------------------------------
void ChromaDiffFlowImageProcessor::GetDescriptions(Params& o_params) const
  {
  o_params["ChromaDiffFlowImageProcessor::ApplyDiffusionFlow()"] = "Applies diffusion flow to the image (denoising)";

  o_params["ChromaFlow_AmountBrightness"] = "Coeficient that controls amount of diffusion. "
    "If you increase it - also increase number of diffusion iterations";
  o_params["ChromaFlow_DiffusionIterations"] = "Number of diffusion iterations. "
    "Does not control diffusion strength (see amount for that). You should not "
    "set this value too small, otherwise diffusion process will be unstable";
  }
//-----------------------------------------------------------------------------
void ChromaDiffFlowImageProcessor::SetParameters(Params& i_params)
  {
  if (i_params.exist("ChromaFlow_AmountBrightness"))
    m_amount_brightness = boost::lexical_cast<float>(i_params.pop("ChromaFlow_AmountBrightness"));
  if (i_params.exist("ChromaFlow_DiffusionIterations"))
    m_diff_num_of_iter = boost::lexical_cast<ulong>(i_params.pop("ChromaFlow_DiffusionIterations"));
  }
//-----------------------------------------------------------------------------
void ChromaDiffFlowImageProcessor::Process(Image& image, const RawImageInfo& image_info, const std::string& method)
  {
  if (method == "ApplyDiffusionFlow()")
    ApplyDiffusionFlow(image, image_info);
  else throw ExImageProcessorError("UnknMethod", "Attempting to call unexisting processing method ChromaDiffFlowImageProcessor::" + method);
  }
//-----------------------------------------------------------------------------
void ChromaDiffFlowImageProcessor::ApplyStep(Image& image, const ImageTmpl<PixelCB>& img_ch,
                                             unsigned row, unsigned col, float step_br, float step_ch)
  {
  PixelCB pix(img_ch.AtBuffer(row, col));

  double m_x = (img_ch.AtBuffer(row, col - 1)[3] - img_ch.AtBuffer(row, col + 1)[3])/2,
         m_y = (img_ch.AtBuffer(row - 1, col)[3] - img_ch.AtBuffer(row, col)[3])/2,
         m_xy = (img_ch.AtBuffer(row - 1, col - 1)[3] - img_ch.AtBuffer(row - 1, col + 1)[3] -
                 img_ch.AtBuffer(row + 1, col - 1)[3] + img_ch.AtBuffer(row + 1, col + 1)[3])/4,
         m_xx = img_ch.AtBuffer(row, col - 1)[3] - 2*img_ch.AtBuffer(row, col)[3]
                 + img_ch.AtBuffer(row, col + 1)[3],
         m_yy = img_ch.AtBuffer(row - 1, col)[3] - 2*img_ch.AtBuffer(row, col)[3]
                + img_ch.AtBuffer(row + 1, col)[3];

  double val = m_xx*m_y*m_y - 2*m_x*m_y*m_xy + m_yy*m_x*m_x;
  if (val < 0)
    pix[3] += -step_ch*pow(-val, 1/3.0) / (1 + sqrt(m_x*m_x + m_y*m_y));
  else
    pix[3] += step_br*pow(val, 1/3.0) / (1 + sqrt(m_x*m_x + m_y*m_y));

  double c_x[3], c_y[3], c_xx[3], c_yy[3];
  c_x[0] = (img_ch.AtBuffer(row, col - 1)[0] - img_ch.AtBuffer(row, col + 1)[0])/2;
  c_x[1] = (img_ch.AtBuffer(row, col - 1)[1] - img_ch.AtBuffer(row, col + 1)[1])/2;
  c_x[2] = (img_ch.AtBuffer(row, col - 1)[2] - img_ch.AtBuffer(row, col + 1)[2])/2;

  c_y[0] = (img_ch.AtBuffer(row - 1, col)[0] - img_ch.AtBuffer(row + 1, col)[0])/2;
  c_y[1] = (img_ch.AtBuffer(row - 1, col)[1] - img_ch.AtBuffer(row + 1, col)[1])/2;
  c_y[2] = (img_ch.AtBuffer(row - 1, col)[2] - img_ch.AtBuffer(row + 1, col)[2])/2;

  c_xx[0] = img_ch.AtBuffer(row, col - 1)[0] - 2*img_ch.AtBuffer(row, col)[0] + img_ch.AtBuffer(row, col + 1)[0];
  c_xx[1] = img_ch.AtBuffer(row, col - 1)[1] - 2*img_ch.AtBuffer(row, col)[1] + img_ch.AtBuffer(row, col + 1)[1];
  c_xx[2] = img_ch.AtBuffer(row, col - 1)[2] - 2*img_ch.AtBuffer(row, col)[2] + img_ch.AtBuffer(row, col + 1)[2];

  c_yy[0] = img_ch.AtBuffer(row - 1, col)[0] - 2*img_ch.AtBuffer(row, col)[0] + img_ch.AtBuffer(row + 1, col)[0];
  c_yy[1] = img_ch.AtBuffer(row - 1, col)[1] - 2*img_ch.AtBuffer(row, col)[1] + img_ch.AtBuffer(row + 1, col)[1];
  c_yy[2] = img_ch.AtBuffer(row - 1, col)[2] - 2*img_ch.AtBuffer(row, col)[2] + img_ch.AtBuffer(row + 1, col)[2];

  double norm = sqrt(c_x[0]*c_x[0] + c_y[0]*c_y[0]
                     + c_x[1]*c_x[1] + c_y[1]*c_y[1]
                     + c_x[2]*c_x[2] + c_y[2]*c_y[2]);
  double dvg[3] = {c_xx[0] + c_yy[0], c_xx[1] + c_yy[1], c_xx[2] + c_yy[2]};

  pix[0] += step_ch*(1/(0.004 + norm)*dvg[0] + pix[0]*norm);
  pix[1] += step_ch*(1/(0.004 + norm)*dvg[1] + pix[1]*norm);
  pix[2] += step_ch*(1/(0.004 + norm)*dvg[2] + pix[2]*norm);

  pix.Normalize(image.GetClampVal());

  Pixel p = pix;

  image.AtBuffer(row, col)[0] = image.ClampChromaPixel(p[0], row, col);
  image.AtBuffer(row, col)[1] = image.ClampChromaPixel(p[1], row, col);
  image.AtBuffer(row, col)[2] = image.ClampChromaPixel(p[2], row, col);
  }
//-----------------------------------------------------------------------------
void ChromaDiffFlowImageProcessor::ApplyDiffusionFlow(Image& image, const RawImageInfo& image_info)
  {
  mp_image_info = &image_info;
  double step_br = m_amount_brightness / m_diff_num_of_iter,
         step_ch = m_amount_chroma / m_diff_num_of_iter;

  ImageTmpl<PixelCB> img_chroma(image.GetWidth(), 0);

  for (ulong num = 0; num < m_diff_num_of_iter; ++num)
    {
    mg_log << INFO << "Applying diffusion flow (pass " << num+1 << " of " << m_diff_num_of_iter << ")" << endlog;

    img_chroma.StartBuffer(0, 5, true);
    for (unsigned row = 0; row < 4; ++row)
      {
      for (unsigned col = 0; col < image.GetWidth(); ++col)
        img_chroma.AtBuffer(row, col) = image.At(row, col);

      img_chroma.FlushBufferRow();
      }

    image.StartBuffer(2, 3);

    for (unsigned row = 2; row < image.GetHeight() - 2; ++row)
      {
      for (unsigned col = 0; col < image.GetWidth(); ++col)
        img_chroma.AtBuffer(row + 2, col) = image.At(row + 2, col);

      for (unsigned col = 2; col < image.GetWidth() - 2; ++col)
        ApplyStep(image, img_chroma, row, col, step_br, step_ch);

      image.FlushBufferRow();
      img_chroma.FlushBufferRow();
      }

    image.StopBuffer();
    img_chroma.StopBuffer();
    }
  }
//-----------------------------------------------------------------------------