/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * HueGradInterpolator - experimental implementation of interpolation based on
 * ``adaptive hues interpolation'' idea. See copyrights.txt for copyright info.
******************************************************************************/

#include "stdafx.h"

#include "HueGradInterpolator.h"
#include "Image.h"
#include "RawImageInfo.h"
//-----------------------------------------------------------------------------
HueGradInterpolator::HueGradInterpolator(): IParametrizable("HueGradInterpolator")
  {
  m_sigma = 1024;
  m_clamp_threshold = m_sigma/2;
  m_max_error = 8192;
  m_fl_smart_clamp = false;
  }
//-----------------------------------------------------------------------------
HueGradInterpolator& HueGradInterpolator::GetInterpolator()
  {
  static HueGradInterpolator hg_int;
  return hg_int;
  }
//-----------------------------------------------------------------------------
void HueGradInterpolator::GetParameters(Params& o_params) const
  {
  o_params["HueGrad_Sigma"] = boost::lexical_cast<std::string>(m_sigma);
  o_params["HueGrad_MaxHueError"] = boost::lexical_cast<std::string>(m_max_error);
  o_params["HueGrad_SmartClamp"] = boost::lexical_cast<std::string>(m_fl_smart_clamp);
  }
//-----------------------------------------------------------------------------
void HueGradInterpolator::GetDescriptions(Params& o_params) const
  {
  o_params["HueGrad_Sigma"] = "Noise deviation";
  o_params["HueGrad_MaxHueError"] = "Maximum allowed apriori hue error";
  o_params["HueGrad_SmartClamp"] = "If true - smart clamping for green will be used";
  }
//-----------------------------------------------------------------------------
void HueGradInterpolator::SetParameters(Params& i_params)
  {
  if (i_params.exist("HueGrad_Sigma"))
    m_sigma = boost::lexical_cast<int>(i_params.pop("HueGrad_Sigma"));
  if (i_params.exist("HueGrad_MaxHueError"))
    m_max_error = boost::lexical_cast<unsigned>(i_params.pop("HueGrad_MaxHueError"));
  if (i_params.exist("HueGrad_SmartClamp"))
    m_fl_smart_clamp = boost::lexical_cast<bool>(i_params.pop("HueGrad_SmartClamp"));
  }
//-----------------------------------------------------------------------------
inline bool HueGradInterpolator::InterpolByHue1(int c1, int c2, int c3, int v1, int v2, int& res)
  {
  if (c1 < m_sigma || c2 < m_sigma || c3 < m_sigma
      || v1 < m_sigma || v2 < m_sigma)
    {
    ++m_num_hue_failed;
    return false; // Otherwise analysis will be unnecessary complicated
    }

  float h0 = (float)v1/c1;
  float h1 = (float)v2/c3;
  float h = (h0 + h1)/2;

  res = (int)(h*c2);

  // Now - let's build error bounds;
  // first let's check upper range ...
  h0 = (float)(v1 + m_sigma)/(c1 - m_sigma);
  h1 = (float)(v2 + m_sigma)/(c3 - m_sigma);
  h = (h0 + h1)/2;

  int max_dev = abs(res - (int)(h*(c2 + m_sigma)));

  // Now - min range ...
  h0 = (float)(v1 - m_sigma)/(c1 + m_sigma);
  h1 = (float)(v2 - m_sigma)/(c3 + m_sigma);
  h = (h0 + h1)/2;

  max_dev = std::max(max_dev, abs(res - (int)(h*(c2 - m_sigma))));

  if (max_dev > m_max_error)
    {
    ++m_num_hue_failed;
    return false;
    }

  ++m_num_hue_ok;
  return true;
  }
//-----------------------------------------------------------------------------
inline bool HueGradInterpolator::InterpolByHue2(int c1, int c2, int c3, int g1, int g2, int& res)
  {
  if (c1 < m_sigma || c2 < m_sigma || c3 < m_sigma
      || g1 < m_sigma || g2 < m_sigma)
    {
    ++m_num_hue_failed;
    return false; // Otherwise analysis will be unnecessary complicated
    }

  int ci0 = c2 + c1;
  int ci1 = c2 + c3;

  float h0 = (float)g1/ci0;
  float h1 = (float)g2/ci1;
  float h = h0 + h1;

  res = (int)(h*c2);

  // Now - let's build error bounds;
  // first let's check upper range ...
  h0 = (float)(g1 + m_sigma)/ci0;
  h1 = (float)(g2 + m_sigma)/ci1;
  h = h0 + h1;

  int max_dev = abs(res - (int)(h*(c2 + m_sigma)));

  // Now - min range ...
  h0 = (float)(g1 - m_sigma)/ci0;
  h1 = (float)(g2 - m_sigma)/ci1;
  h = h0 + h1;

  max_dev = std::max(max_dev, abs(res - (int)(h*(c2 - m_sigma))));

  if (max_dev > m_max_error)
    {
    ++m_num_hue_failed;
    return false;
    }

  ++m_num_hue_ok;
  return true;
  }
//-----------------------------------------------------------------------------
inline uchar HueGradInterpolator::FillGradientsForGreen(const Image& image, int CC[], int P[], unsigned row, unsigned col)
  {
  uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);
  assert (ind != 1);

  uchar ind1 = (ind+2)%4;

  CC[0] = 2*(int)image.At(row, col)[ind]
	  - (int)image.At(row, col - 2)[ind]
	  - (int)image.At(row, col + 2)[ind];
  if (!InterpolByHue2((int)image.At(row, col - 2)[ind], (int)image.At(row, col)[ind],
		      (int)image.At(row, col + 2)[ind], (int)image.At(row, col - 1)[1],
		      (int)image.At(row, col + 1)[1], P[0]))
    P[0] = ((int)image.At(row, col - 1)[1] + (int)image.At(row, col + 1)[1])/2
	   + CC[0] / 4;
  CC[0] = abs(CC[0]) + abs((int)image.At(row, col - 1)[1] - (int)image.At(row, col + 1)[1]);
  
  CC[1] = 2*(int)image.At(row, col)[ind]
	  - (int)image.At(row - 2, col)[ind]
	  - (int)image.At(row + 2, col)[ind];
  if (!InterpolByHue2((int)image.At(row - 2, col)[ind], (int)image.At(row, col)[ind],
		      (int)image.At(row + 2, col)[ind], (int)image.At(row - 1, col)[1],
		      (int)image.At(row + 1, col)[1], P[1]))
    P[1] = ((int)image.At(row - 1, col)[1] + (int)image.At(row + 1, col)[1]) / 2
	   + CC[1]/4;
  CC[1] = abs(CC[1]) + abs((int)image.At(row - 1, col)[1] - (int)image.At(row + 1, col)[1]);

  if (CC[0] <= CC[1])
    return 0;
  else
    return 1;
  }
//-----------------------------------------------------------------------------
inline void HueGradInterpolator::InterpolateGreenPixel(Image& image, unsigned row, unsigned col)
  {
  int CC[2], P[2];

  uchar ind_c = FillGradientsForGreen(image, CC, P, row, col);

  if (m_fl_smart_clamp)
    {
    ushort gc = image.ClampGreenPixel(P[ind_c], row, col),
	   g = image.ClampChromaPixel(P[ind_c], row, col);

    if (gc == g)
      image.At(row, col)[1] = g;
    else if (abs(gc - g) < m_clamp_threshold)
      image.At(row, col)[1] = gc;
    else if (gc < g)
      image.At(row, col)[1] = g - m_clamp_threshold;
    else
      image.At(row, col)[1] = g + m_clamp_threshold;
    }
  else
    image.At(row, col)[1] = image.ClampGreenPixel(P[ind_c], row, col);
  }
//-----------------------------------------------------------------------------
inline void HueGradInterpolator::InterpolateChromaPixelPass1(Image& image, unsigned row, unsigned col)
  {
  uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);
  assert (ind != 1);

  uchar ind1 = (ind+2)%4;

  int CC[2], P[2];

  CC[0] =  2*(int)image.At(row, col)[1]
	   - (int)image.At(row - 1, col + 1)[1]
	   - (int)image.At(row + 1, col - 1)[1];
  if (!InterpolByHue1((int)image.At(row - 1, col + 1)[1], (int)image.At(row, col)[1],
		      (int)image.At(row + 1, col - 1)[1], (int)image.At(row - 1, col + 1)[ind1],
		      (int)image.At(row + 1, col - 1)[ind1], P[0]))
    P[0] = ((int)image.At(row - 1, col + 1)[ind1]
	    + (int)image.At(row + 1, col - 1)[ind1])/2
	   + CC[0] / 2;
  CC[0] = abs(CC[0])
	  + abs((int)image.At(row - 1, col + 1)[ind1]
		- (int)image.At(row + 1, col - 1)[ind1]);


  CC[1] = 2*(int)image.At(row, col)[1]
	  - (int)image.At(row - 1, col - 1)[1]
	  - (int)image.At(row + 1, col + 1)[1];
  if (!InterpolByHue1((int)image.At(row - 1, col - 1)[1], (int)image.At(row, col)[1],
		      (int)image.At(row + 1, col + 1)[1], (int)image.At(row - 1, col - 1)[ind1],
		      (int)image.At(row + 1, col + 1)[ind1], P[1]))
    P[1] = ((int)image.At(row - 1, col - 1)[ind1]
	    + (int)image.At(row + 1, col + 1)[ind1])/2
	   + CC[1] / 2;
  CC[1] = abs(CC[1])
	  + abs((int)image.At(row - 1, col - 1)[ind1]
		- (int)image.At(row + 1, col + 1)[ind1]);

  int color;

  if (CC[0] <= CC[1])
    color = P[0];
  else
    color = P[1];

  image.At(row, col)[ind1] = image.ClampChromaPixel(color, row, col);
  }
//-----------------------------------------------------------------------------
inline void HueGradInterpolator::InterpolateChromaPixelPass2(Image& image, unsigned row, unsigned col, uchar ind)
  {
  assert (mp_image_info -> GetCurrentColorIndex(row, col) == 1);
  ushort ind1 = (ind+2)%4;

  int CC[2], P[2];

  CC[0] =  2*(int)image.At(row, col)[1]
	   - (int)image.At(row - 1, col)[1]
	   - (int)image.At(row + 1, col)[1];
  if (!InterpolByHue1((int)image.At(row - 1, col)[1], (int)image.At(row, col)[1],
		      (int)image.At(row + 1, col)[1], (int)image.At(row - 1, col)[ind],
		      (int)image.At(row + 1, col)[ind], P[0]))
    P[0] = ((int)image.At(row - 1, col)[ind]
	    + (int)image.At(row + 1, col)[ind])/2
	   + CC[0] / 2;
  CC[0] = abs(CC[0])
	  + abs((int)image.At(row - 1, col)[ind]
		- (int)image.At(row + 1, col)[ind]);

  CC[1] = 2*(int)image.At(row, col)[1]
	  - (int)image.At(row, col - 1)[1]
	  - (int)image.At(row, col + 1)[1];
  if (!InterpolByHue1((int)image.At(row, col - 1)[1], (int)image.At(row, col)[1],
		      (int)image.At(row, col + 1)[1], (int)image.At(row, col - 1)[ind],
		      (int)image.At(row, col + 1)[ind], P[1]))
    P[1] = ((int)image.At(row, col - 1)[ind]
	    + (int)image.At(row, col + 1)[ind])/2
	   + CC[1] / 2;
  CC[1] = abs(CC[1])
	  + abs((int)image.At(row, col - 1)[ind]
		- (int)image.At(row, col + 1)[ind]);

  int color;
  if (CC[0] <= CC[1])
    color = P[0];
  else
    color = P[1];

  image.At(row, col)[ind] = image.ClampChromaPixel(color, row, col);
  }
//-----------------------------------------------------------------------------
bool HueGradInterpolator::Interpolate(Image& image, const RawImageInfo& image_info)
  {
  mg_log << INFO << "Running Hue-Gradient interpolation" << endlog;
  mp_image_info = &image_info;

  m_num_hue_failed = m_num_hue_ok = 0;

  mg_log << INFO << "Interpolating green" << endlog;

  for (unsigned row = 2; row < image.GetHeight() - 2; ++row)
    for (unsigned col = 2; col < image.GetWidth() - 2; ++col)
      {
      uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);
      if (ind != 1)
	InterpolateGreenPixel(image, row, col);
      }

  mg_log << DEBUG << "Failed hues: " << m_num_hue_failed << ", processed hues: " << m_num_hue_ok << endlog;

  m_num_hue_failed = m_num_hue_ok = 0;

  image.SetState(Image::FULL_GREEN_COLORS);

  mg_log << INFO << "Interpolating chroma (1st pass)" << endlog;

  for (unsigned row = 2; row < image.GetHeight() - 2; ++row)
    for (unsigned col = 2; col < image.GetWidth() - 2; ++col)
      {
      uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);

      if (ind != 1)
	InterpolateChromaPixelPass1(image, row, col);
      }

  mg_log << DEBUG << "Failed hues: " << m_num_hue_failed << ", processed hues: " << m_num_hue_ok << endlog;

  m_num_hue_failed = m_num_hue_ok = 0;

  mg_log << INFO << "Interpolating chroma (2nd pass)" << endlog;

  for (unsigned row = 2; row < image.GetHeight() - 2; ++row)
    for (unsigned col = 2; col < image.GetWidth() - 2; ++col)
      {
      uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);

      if (ind == 1)
	{
	InterpolateChromaPixelPass2(image, row, col, 0);
	InterpolateChromaPixelPass2(image, row, col, 2);
	}
      }

  mg_log << DEBUG << "Failed hues: " << m_num_hue_failed << ", processed hues: " << m_num_hue_ok << endlog;

  image.SetState(Image::FULL_COLORS);

  mg_log << TRACE << "Finished" << endlog;

  return true;
  }
//-----------------------------------------------------------------------------
