/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#ifndef IMAGE_H
#define IMAGE_H

#include "Pixel.h"
#include "Position.h"
//-----------------------------------------------------------------------------
class RawImageInfo;
//-----------------------------------------------------------------------------
// General Image class (almost matrix)
//
//
template <class PixelType>
class ImageTmpl
  {
  public:
    ImageTmpl();
    ImageTmpl(unsigned width, unsigned height, unsigned trim_width = 0);
    ImageTmpl(const ImageTmpl<PixelType>& image);

    void Create(unsigned width, unsigned height, unsigned trim_width = 0);
    void Clear();

    void Set(const PixelType& val);

    // general access operations
    unsigned GetWidth() const;
    unsigned GetHeight() const;
    unsigned GetTrimWidth() const;

    const PixelType& At(unsigned row, unsigned col) const;
    PixelType& At(unsigned row, unsigned col);

    const PixelType& At(const Position& pos) const;
    PixelType& At(const Position& pos);

    std::vector<PixelType>& GetContainer();
    const std::vector<PixelType>& GetContainer() const;

    ImageTmpl<PixelType>& operator =(const ImageTmpl<PixelType>& image);

    // Buffer functionality
    void StartBuffer(unsigned start_row, unsigned height, bool fl_dummy_buffer = false);

    const PixelType& AtBuffer(unsigned row, unsigned col) const;
    PixelType& AtBuffer(unsigned row, unsigned col);

    const PixelType& AtBuffer(const Position& pos) const;
    PixelType& AtBuffer(const Position& pos);

    void FlushBufferRow();

    void StopBuffer();

  protected:
    std::vector<PixelType> m_image;
    unsigned m_width, m_height;
    unsigned m_trim_width;
    std::vector<PixelType> m_buffer;
    std::vector<ulong> m_buffer_row_pos;
    unsigned m_buffer_height, m_buffer_start, m_buffer_offs;
    bool m_fl_dummy_buffer;
  };
//-----------------------------------------------------------------------------
typedef ImageTmpl<PixelCB> ImageCB;
//-----------------------------------------------------------------------------
// Extended image class for storage of RGB 16bit-per-sample images
//
class Image: public ImageTmpl<Pixel>
  {
  public:
    enum ImageState // Used to specify what information is already
		    // available in image; after loading it should be RAW_SAMPLES,
		    // after interpolation of green - FULL_GREEN_COLORS,
		    // and after interpolation is complete - FULL_COLORS
      {
      UNKNOWN,
      RAW_SAMPLES,
      FULL_GREEN_COLORS,
      FULL_COLORS
      };

  public:
    Image();
    Image(unsigned width, unsigned height, ushort clamp_val = 0xFFFF, unsigned trim_width = 0);
    Image(const Image& image);

    void Create(unsigned width, unsigned height, ushort clamp_val = 0xFFFF, unsigned trim_width = 0);
    void Clear();

    void Zero();

    // general access operations
    const ushort GetClampVal() const;

    const ImageState GetState() const;

    void SetState(ImageState state);

    const std::string& GetProfile() const;

    void SetProfile(const std::string& profile);

    Image& operator =(const Image& image);

    ushort ClampGreenPixel(int val, unsigned row, unsigned col) const;

    ushort ClampChromaPixel(int val, unsigned row, unsigned col) const;

    unsigned GetIntensity(unsigned row, unsigned col) const;

  private:
    ushort m_clamp_val;
    ImageState m_state;
    std::string m_profile;
  };
//-----------------------------------------------------------------------------
// ImageTmpl inlines implementation:
//-----------------------------------------------------------------------------
template <class PixelType>
inline unsigned ImageTmpl<PixelType>::GetWidth() const
  {
  return m_width;
  }
//-----------------------------------------------------------------------------
template <class PixelType>
inline unsigned ImageTmpl<PixelType>::GetHeight() const
  {
  return m_height;
  }
//-----------------------------------------------------------------------------
template <class PixelType>
inline unsigned ImageTmpl<PixelType>::GetTrimWidth() const
  {
  return m_trim_width;
  }
//-----------------------------------------------------------------------------
template <class PixelType>
inline const PixelType& ImageTmpl<PixelType>::At(unsigned row, unsigned col) const
  {
  assert(row < m_height && col < m_width && (m_buffer.empty() || row >= m_buffer_start));
  return m_image[row*m_width + col];
  }
//-----------------------------------------------------------------------------
template <class PixelType>
inline PixelType& ImageTmpl<PixelType>::At(unsigned row, unsigned col)
  {
  assert(row < m_height && col < m_width && (m_buffer.empty() || row >= m_buffer_start));
  return m_image[row*m_width + col];
  }
//-----------------------------------------------------------------------------
template <class PixelType>
inline const PixelType& ImageTmpl<PixelType>::At(const Position& pos) const
  {
  return At(pos.m_row, pos.m_col);
  }
//-----------------------------------------------------------------------------
template <class PixelType>
inline PixelType& ImageTmpl<PixelType>::At(const Position& pos)
  {
  return At(pos.m_row, pos.m_col);
  }
//-----------------------------------------------------------------------------
template <class PixelType>
inline std::vector<PixelType>& ImageTmpl<PixelType>::GetContainer()
  {
  return m_image;
  }
//-----------------------------------------------------------------------------
template <class PixelType>
inline const std::vector<PixelType>& ImageTmpl<PixelType>::GetContainer() const
  {
  return m_image;
  }
//-----------------------------------------------------------------------------
template <class PixelType>
inline const PixelType& ImageTmpl<PixelType>::AtBuffer(unsigned row, unsigned col) const
  {
  assert((row < m_height || m_fl_dummy_buffer) && col < m_width
	 && row >= m_buffer_start && row < m_buffer_start + m_buffer_height);
  return m_buffer[m_buffer_row_pos[row - m_buffer_start] + col];
  }
//-----------------------------------------------------------------------------
template <class PixelType>
inline PixelType& ImageTmpl<PixelType>::AtBuffer(unsigned row, unsigned col)
  {
  assert((row < m_height || m_fl_dummy_buffer) && col < m_width
	 && row >= m_buffer_start && row < m_buffer_start + m_buffer_height);
  return m_buffer[m_buffer_row_pos[row - m_buffer_start] + col];
  }
//-----------------------------------------------------------------------------
template <class PixelType>
inline const PixelType& ImageTmpl<PixelType>::AtBuffer(const Position& pos) const
  {
  return AtBuffer(pos.m_row, pos.m_col);
  }
//-----------------------------------------------------------------------------
template <class PixelType>
inline PixelType& ImageTmpl<PixelType>::AtBuffer(const Position& pos)
  {
  return AtBuffer(pos.m_row, pos.m_col);
  }
//-----------------------------------------------------------------------------
// Image inlines implementation:
//-----------------------------------------------------------------------------
inline const ushort Image::GetClampVal() const
  {
  return m_clamp_val;
  }
//-----------------------------------------------------------------------------
inline const Image::ImageState Image::GetState() const
  {
  return m_state;
  }
//-----------------------------------------------------------------------------
inline void Image::SetState(ImageState state)
  {
  m_state = state;
  }
//-----------------------------------------------------------------------------
inline ushort Image::ClampGreenPixel(int val, unsigned row, unsigned col) const
  {
  if (val < 0)
    val = 0;

  if (val > GetClampVal())
    val = GetClampVal();

  ushort cur_max = At(row - 1, col)[1], cur_min = cur_max, v;

  if ((v = At(row + 1, col)[1]) > cur_max)
    cur_max = v;
  if (v < cur_min)
    cur_min = v;

  if ((v = At(row, col - 1)[1]) > cur_max)
    cur_max = v;
  if (v < cur_min)
    cur_min = v;

  if ((v = At(row, col + 1)[1]) > cur_max)
    cur_max = v;
  if (v < cur_min)
    cur_min = v;

  if (cur_max < val)
    val = cur_max;
  if (cur_min > val)
    val = cur_min;

  return val;
  }
//-----------------------------------------------------------------------------
inline ushort Image::ClampChromaPixel(int val, unsigned row, unsigned col) const
  {
  if (val < 0)
    val = 0;
  else if (val > GetClampVal())
    val = GetClampVal();

  return val;
  }
//-----------------------------------------------------------------------------
inline unsigned Image::GetIntensity(unsigned row, unsigned col) const
  {
  return (unsigned)sqrt((float)At(row, col)[0]*(float)At(row, col)[0] +
			(float)At(row, col)[1]*(float)At(row, col)[1] +
			(float)At(row, col)[2]*(float)At(row, col)[2]);
  }
//-----------------------------------------------------------------------------
inline const std::string& Image::GetProfile() const
  {
  return m_profile;
  }
//-----------------------------------------------------------------------------
inline void Image::SetProfile(const std::string& profile)
  {
  m_profile = profile;
  }
//-----------------------------------------------------------------------------
#endif // IMAGE_H
