/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * ImageSimpleFilters - simple algorithms for image filtering.
******************************************************************************/

#include "stdafx.h"

#include "ImageSimpleFilters.h"
#include "ImageUtilities.h"
#include "Image.h"
#include "RawImageInfo.h"
//-----------------------------------------------------------------------------
ImageSimpleFilters::ImageSimpleFilters(): IParametrizable("ImageSimpleFilters")
  {
  m_median_dim = 1;
  m_raw_median_dim = 2;
  m_median_strength = 1.0;
  m_noise_filter_strength = 0.8;
  }
//-----------------------------------------------------------------------------
ImageSimpleFilters& ImageSimpleFilters::GetImageProcessor()
  {
  static ImageSimpleFilters filters;
  return filters;
  }
//-----------------------------------------------------------------------------
void ImageSimpleFilters::GetParameters(Params& o_params) const
  {
  o_params["ImageSimpleFilters_MedianWidth"] = boost::lexical_cast<std::string>(m_median_dim);
  o_params["ImageSimpleFilters_RawMedianWidth"] = boost::lexical_cast<std::string>(m_raw_median_dim);
  o_params["ImageSimpleFilters_MedianStrength"] = boost::lexical_cast<std::string>(m_median_strength);
  o_params["ImageSimpleFilters_NoiseFilterStrength"] = boost::lexical_cast<std::string>(m_noise_filter_strength);
  }
//-----------------------------------------------------------------------------
void ImageSimpleFilters::GetDescriptions(Params& o_params) const
  {
  o_params["ImageSimpleFilters::Median()"] = "Simple square median filter";
  o_params["ImageSimpleFilters::DiagonalMedian()"] = "Simple diagonal median filter";
  o_params["ImageSimpleFilters::UltraResolution()"] = "Ultra resolution filter";
  o_params["ImageSimpleFilters::SimpleNoiseFilter()"] = "Simple noise filter (not very useful)";

  o_params["ImageSimpleFilters_MedianWidth"] = "Width of region that is used by median filter";
  o_params["ImageSimpleFilters_RawMedianWidth"] = "Width of region that is used by median filter when filtering in raw mode";
  o_params["ImageSimpleFilters_MedianStrength"] = "Strength of median filter (should be between 0.0 and 1.0)";
  o_params["ImageSimpleFilters_NoiseFilterStrength"] = "Strength of noise filter (should be between 0.0 and 1.0)";;
  }
//-----------------------------------------------------------------------------
void ImageSimpleFilters::SetParameters(Params& i_params)
  {
  if (i_params.exist("ImageSimpleFilters_MedianWidth"))
    m_median_dim = boost::lexical_cast<ulong>(i_params.pop("ImageSimpleFilters_MedianWidth"));
  if (i_params.exist("ImageSimpleFilters_RawMedianWidth"))
    m_raw_median_dim = boost::lexical_cast<ulong>(i_params.pop("ImageSimpleFilters_RawMedianWidth"));
  if (i_params.exist("ImageSimpleFilters_MedianStrength"))
    m_median_strength = boost::lexical_cast<float>(i_params.pop("ImageSimpleFilters_MedianStrength"));
  if (i_params.exist("ImageSimpleFilters_NoiseFilterStrength"))
    m_noise_filter_strength = boost::lexical_cast<float>(i_params.pop("ImageSimpleFilters_NoiseFilterStrength"));
  }
//-----------------------------------------------------------------------------
void ImageSimpleFilters::Process(Image& image, const RawImageInfo& image_info, const std::string& method)
  {
  if (method == "Median()")
    FilterByMedian(image, image_info, false);
  else if (method == "DiagonalMedian()")
    FilterByMedian(image, image_info, true);
  else if (method == "UltraResolution()")
    ApplyUltraResolution(image, image_info);
  else if (method == "SimpleNoiseFilter()")
    SimpleNoiseFilter(image, image_info);
  else throw ExImageProcessorError("UnknMethod", "Attempting to call unexisting processing method ImageSimpleFilters::" + method);
  }
//-----------------------------------------------------------------------------
void ImageSimpleFilters::RawFilterPixelByMedian(Image& image, unsigned row, unsigned col,
						uchar ind, std::vector<ushort>& arr)
  {
  uchar i = 0;
  for (unsigned r = row - m_raw_median_dim; r <= row + m_raw_median_dim; r += 2)
    for (unsigned c = col - m_raw_median_dim; c <= col + m_raw_median_dim; c += 2)
      arr[i++] = image.At(r, c)[ind];

  std::sort(arr.begin(), arr.end());

  image.AtBuffer(row, col)[ind] = (ushort)((m_median_strength - 1)*image.At(row, col)[ind]
    + m_median_strength*arr[arr.size()/2]);
  }
//-----------------------------------------------------------------------------
void ImageSimpleFilters::FilterPixelByMedian(Image& image, unsigned row, unsigned col,
					     uchar ind, std::vector<ushort>& arr)
  {
  uchar i = 0;
  for (unsigned r = row - m_median_dim; r <= row + m_median_dim; ++r)
    for (unsigned c = col - m_median_dim; c <= col + m_median_dim; ++c)
      arr[i++] = image.At(r, c)[ind];

  std::sort(arr.begin(), arr.end());

  image.AtBuffer(row, col)[ind] = (ushort)((m_median_strength - 1)*image.At(row, col)[ind]
    + m_median_strength*arr[arr.size()/2]);
  }
//-----------------------------------------------------------------------------
void ImageSimpleFilters::RawFilterPixelByDiagMedian(Image& image, unsigned row, unsigned col,
						    uchar ind, std::vector<ushort>& arr)
  {
  uchar i = 0;
  for (unsigned r = row - m_raw_median_dim; r <= row + m_raw_median_dim; r += 2)
    arr[i++] = image.At(r, col)[ind];

  for (unsigned c = col - m_raw_median_dim; c <= col + m_raw_median_dim; c += 2)
    if (c != col)
      arr[i++] = image.At(row, c)[ind];

  std::sort(arr.begin(), arr.end());

  image.AtBuffer(row, col)[ind] = (ushort)((m_median_strength - 1)*image.At(row, col)[ind]
    + m_median_strength*arr[arr.size()/2]);
  }
//-----------------------------------------------------------------------------
void ImageSimpleFilters::FilterPixelByDiagMedian(Image& image, unsigned row, unsigned col,
						 uchar ind, std::vector<ushort>& arr)
  {
  uchar i = 0;
  for (unsigned r = row - m_median_dim; r <= row + m_median_dim; ++r)
    arr[i++] = image.At(r, col)[ind];

  for (unsigned c = col - m_median_dim; c <= col + m_median_dim; ++c)
    if (c != col)
      arr[i++] = image.At(row, c)[ind];

  std::sort(arr.begin(), arr.end());

  image.AtBuffer(row, col)[ind] = (ushort)((m_median_strength - 1)*image.At(row, col)[ind]
    + m_median_strength*arr[arr.size()/2]);
  }
//-----------------------------------------------------------------------------
void ImageSimpleFilters::FilterByMedian(Image& image, const RawImageInfo& image_info,
					bool fl_use_diag)
  {
  ImageUtilities::GetImageProcessor().ConvertRGB2RGGB(image);

  mg_log << INFO << "Filtering by median" << endlog;

  if (image.GetState() == Image::FULL_GREEN_COLORS)
    {
//    image.StartBuffer(m_raw_median_dim, 1 + 2*m_raw_median_dim);
    image.StartBuffer(m_raw_median_dim, 1 + m_raw_median_dim);
    ulong size;
    if (fl_use_diag)
      size = 1 + 2*m_raw_median_dim;
    else
      size = (1 + m_raw_median_dim)*(1 + m_raw_median_dim);
    std::vector<ushort> arr(size);

    uchar ind;
    for (unsigned row = m_raw_median_dim; row < image.GetHeight() - m_raw_median_dim; ++row)
      {
      for (unsigned col = m_raw_median_dim; col < image.GetWidth() - m_raw_median_dim; ++col)
	if ((ind = image_info.GetCurrentColorIndex(row, col)) != 1)
	  if (fl_use_diag)
	    RawFilterPixelByDiagMedian(image, row, col, ind, arr);
	  else
	    RawFilterPixelByMedian(image, row, col, ind, arr);

      image.FlushBufferRow();
      }
    }
  else
    {
//    image.StartBuffer(m_median_dim, 1 + 2*m_median_dim);
    image.StartBuffer(m_median_dim, 1 + m_median_dim);
    ulong size;
    if (fl_use_diag)
      size = 1 + 4*m_median_dim;
    else
      size = (1 + 2*m_median_dim)*(1 + 2*m_median_dim);
    std::vector<ushort> arr(size);

    for (unsigned row = m_median_dim; row < image.GetHeight() - m_median_dim; ++row)
      {
      for (unsigned col = m_median_dim; col < image.GetWidth() - m_median_dim; ++col)
	for (uchar ind = 0; ind < 3; ind += 2)
	  if (fl_use_diag)
	    FilterPixelByDiagMedian(image, row, col, ind, arr);
	  else
	    FilterPixelByMedian(image, row, col, ind, arr);

      image.FlushBufferRow();
      }
    }
  image.StopBuffer();

  ImageUtilities::GetImageProcessor().ConvertRGGB2RGB(image);
  }
//-----------------------------------------------------------------------------
void ImageSimpleFilters::ApplyUltraResolution(Image& image, const RawImageInfo& image_info)
  {
  mg_log << INFO << "Applying ultra resolution" << endlog;
  image.StartBuffer(2, 3);
  
  for (unsigned row = 2; row < image.GetHeight() - 2; ++row)
    {
    for (unsigned col = 2; col < image.GetWidth() - 2; ++col)
      {
      float avG = 2*((int)image.At(row - 2, col)[1] +
		     (int)image.At(row + 2, col)[1] +
		     (int)image.At(row, col - 2)[1] + 
		     (int)image.At(row, col + 2)[1]) +

		(int)image.At(row - 2, col - 2)[1] +
		(int)image.At(row - 2, col + 2)[1] +
		(int)image.At(row + 2, col - 2)[1] + 
		(int)image.At(row + 2, col + 2)[1];

      for (uchar ind = 0; ind < 3; ind += 2)
	{
	if (image_info.GetCurrentColorIndex(row, col) != ind)
	  continue;

	float avC = 2*((int)image.At(row - 2, col)[ind] +
		       (int)image.At(row + 2, col)[ind] +
		       (int)image.At(row, col - 2)[ind] + 
		       (int)image.At(row, col + 2)[ind]) +

		(int)image.At(row - 2, col - 2)[ind] +
		(int)image.At(row - 2, col + 2)[ind] +
		(int)image.At(row + 2, col - 2)[ind] + 
		(int)image.At(row + 2, col + 2)[ind];

	if (avG < avC*8)
	  image.AtBuffer(row, col)[1] = image.ClampChromaPixel((int)(avG/avC*image.At(row, col)[ind]), row, col);
	}
      }
    image.FlushBufferRow();
    }

  image.StopBuffer();
  }
//-----------------------------------------------------------------------------
void ImageSimpleFilters::SimpleNoiseFilter(Image& image, const RawImageInfo& image_info)
  {
  mg_log << INFO << "Applying simple noise filter" << endlog;
  image.StartBuffer(2, 3);

  uchar ind;
  for (unsigned row = 2; row < image.GetHeight() - 2; ++row)
    {
    for (unsigned col = 2; col < image.GetWidth() - 2; ++col)
      if ((ind = image_info.GetCurrentColorIndex(row, col)) == 1)
	{
	ushort cur_max = image.At(row - 1, col - 1)[1], cur_min = cur_max, v;
	if ((v = image.At(row + 1, col - 1)[1]) > cur_max)
	  cur_max = v;
	if (v < cur_min)
	  cur_min = v;
	if ((v = image.At(row - 1, col + 1)[1]) > cur_max)
	  cur_max = v;
	if (v < cur_min)
	  cur_min = v;
	if ((v = image.At(row + 1, col + 1)[1]) > cur_max)
	  cur_max = v;
	if (v < cur_min)
	  cur_min = v;

	if (cur_max < m_noise_filter_strength*image.At(row, col)[1])
	  image.AtBuffer(row, col)[1] = cur_max;
	if (m_noise_filter_strength*cur_min > image.At(row, col)[1])
	  image.AtBuffer(row, col)[1] = cur_min;
	}
      else
	{
	ushort cur_max = image.At(row - 2, col)[ind], cur_min = cur_max, v;
	if ((v = image.At(row + 2, col)[ind]) > cur_max)
	  cur_max = v;
	if (v < cur_min)
	  cur_min = v;
	if ((v = image.At(row, col - 2)[ind]) > cur_max)
	  cur_max = v;
	if (v < cur_min)
	  cur_min = v;
	if ((v = image.At(row, col + 2)[ind]) > cur_max)
	  cur_max = v;
	if (v < cur_min)
	  cur_min = v;

	if (cur_max < m_noise_filter_strength*image.At(row, col)[ind])
	  image.AtBuffer(row, col)[ind] = cur_max;
	if (m_noise_filter_strength*cur_min > image.At(row, col)[ind])
	  image.AtBuffer(row, col)[ind] = cur_min;
	}
    image.FlushBufferRow();
    }

  image.StopBuffer();
  }
//-----------------------------------------------------------------------------
