/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * ImageTransfer - implementation of reading/writing image files with/without
 * ImageMagick support and conversion of images between internal format and
 * ImageMagick format (in fact, without ImageMagick support only writing and
 * only PSD file format is supported).
******************************************************************************/

#include "stdafx.h"

// We will need htonl(), htons(), etc
//
#ifdef WIN32
#include <winsock.h>
#else
#include <sys/param.h>
#endif

#include "ImageTransfer.h"
#include "Image.h"
//-----------------------------------------------------------------------------
ImageTransfer::ImageTransfer(): IParametrizable("ImageTransfer")
  {
  m_add_trim = 0;
  m_format = "by-extention";
#ifdef USE_IMAGE_MAGICK
  m_quality = 90;
#endif // USE_IMAGE_MAGICK
  }
//-----------------------------------------------------------------------------
ImageTransfer& ImageTransfer::GetTransfer()
  {
  static ImageTransfer transfer;
  return transfer;
  }
//-----------------------------------------------------------------------------
void ImageTransfer::GetParameters(Params& o_params) const
  {
  o_params["ImageTransfer_AdditionalTrimWidth"] = boost::lexical_cast<std::string>(m_add_trim);
  o_params["ImageTransfer_Format"] = boost::lexical_cast<std::string>(m_format);

#ifdef USE_IMAGE_MAGICK
  o_params["ImageTransfer_Quality"] = boost::lexical_cast<std::string>(m_quality);
#endif // USE_IMAGE_MAGICK
  }
//-----------------------------------------------------------------------------
void ImageTransfer::GetDescriptions(Params& o_params) const
  {
  o_params["ImageTransfer_AdditionalTrimWidth"] = "The width of trim border "
    "in result picture (you may leave 0, but pixels very near the border usually "
    "have not very good quality, so at least 2 is recommended)";
  o_params["ImageTransfer_Format"] = "Default image format for output "
#ifndef USE_IMAGE_MAGICK
    "(only PSD is supported in this version; use version compiled with ImageMagick "
    "support to output images in additional formats)";
#else
    "(see ImageMagick documentation for the complete list of supported formats; "
    "most probably every format you might need is supported)";
#endif

#ifdef USE_IMAGE_MAGICK
    o_params["ImageTransfer_Quality"] = "Output quality setting for file formats that support it - "
    "f.e. JPEG; should be from 0 to 100";
#endif
  }
//-----------------------------------------------------------------------------
void ImageTransfer::SetParameters(Params& i_params)
  {
  if (i_params.exist("ImageTransfer_AdditionalTrimWidth"))
    m_add_trim = boost::lexical_cast<int>(i_params.pop("ImageTransfer_AdditionalTrimWidth"));
  if (i_params.exist("ImageTransfer_Format"))
    m_format = boost::lexical_cast<std::string>(i_params.pop("ImageTransfer_Format"));

#ifdef USE_IMAGE_MAGICK
  if (i_params.exist("ImageTransfer_Quality"))
    m_quality = boost::lexical_cast<ushort>(i_params.pop("ImageTransfer_Quality"));
#endif // USE_IMAGE_MAGICK
  }
//-----------------------------------------------------------------------------
void ImageTransfer::GetSupportedFormats(std::list<std::string>& formats) const
  {
#ifndef USE_IMAGE_MAGICK
  formats.push_back("PSD (write only) - Adobe Photoshop bitmap");
#else
  Magick::ExceptionInfo exception;
  Magick::GetExceptionInfo(&exception);
    
  const Magick::MagickInfo* info = Magick::GetMagickInfo("*", &exception);

  for (; info; info = info -> next)
    {
    if (!info -> description || !info -> name)
      continue;

    if (info -> decoder && info -> decoder)
      formats.push_back((std::string)info -> name + " - " + info -> description);
    else if (!info -> decoder)
      formats.push_back((std::string)info -> name + " (write only) - " + info -> description);
    else if (!info -> encoder)
      formats.push_back((std::string)info -> name + " (read only) - " + info -> description);
    }
#endif
  }
//-----------------------------------------------------------------------------
void ImageTransfer::Read(const std::string& file_name, Image& image)
  {
#ifndef USE_IMAGE_MAGICK
  throw ExImageTransferError("NoImageMagick", "This version was compiled without ImageMagick support; cannot read image files");
#else
  Magick::Image* im_image = Read(file_name);

  IMImageToImage(image, im_image);

  Magick::DestroyImage(im_image);
#endif // #ifndef USE_IMAGE_MAGICK
  }
//-----------------------------------------------------------------------------
void ImageTransfer::Write(const std::string& file_name, const Image& image)
  {
#ifdef USE_IMAGE_MAGICK
  Magick::Image* im_image = ImageToIMImage(image, true);

  Write(file_name, im_image);

  Magick::DestroyImage(im_image);
#else
  WritePSD(file_name, image);
#endif // #ifdef USE_IMAGE_MAGICK
  }
//-----------------------------------------------------------------------------
void ImageTransfer::WritePSD(const std::string& file_name, const Image& image)
  {
  uchar head[] = {
    '8','B','P','S',            /* signature */
    0,1,0,0,0,0,0,0,            /* version and reserved */
    0,3,                        /* number of channels */
    0,0,0,0,                    /* height, big-endian */
    0,0,0,0,                    /* width, big-endian */
    0,16,                       /* 16-bit color */
    0,3,                        /* mode (1=grey, 3=rgb) */
    0,0,0,0,                    /* color mode data */
    0,0,0,0,                    /* image resources */
    0,0,0,0,                    /* layer/mask info */
    0,0                         /* no compression */
  };

  mg_log << TRACE << "Writing PSD file " << file_name << endlog;

  FILE* ofp = fopen(file_name.c_str(), "wb");

  unsigned trim = (int)image.GetTrimWidth() + m_add_trim;

  long hh = htonl(image.GetHeight() - 2*trim),
       hw = htonl(image.GetWidth() - 2*trim);

  memcpy (head + 14, &hh, 4);
  memcpy (head + 18, &hw, 4);

  fwrite (head, 40, 1, ofp);

  ushort* buf = new ushort[image.GetWidth() - 2*trim];

  for (uchar ind = 0; ind < 3; ++ind)
    for (unsigned row = trim; row < image.GetHeight() - trim; ++row)
      {
      for (unsigned col = trim; col < image.GetWidth() - trim; ++col)
	buf[col - trim] = htons(image.At(row, col)[ind]);

      fwrite(buf, image.GetWidth() - 2*trim, 2, ofp);
      }

  delete buf;
  fclose(ofp);

  mg_log << TRACE << "Finished" << endlog;
  }
//-----------------------------------------------------------------------------
#ifdef USE_IMAGE_MAGICK
//
Magick::Image* ImageTransfer::Read(const std::string& file_name)
  {
  Magick::ExceptionInfo exception;

  Magick::Image *im_image;
  Magick::ImageInfo *im_image_info;

  Magick::GetExceptionInfo(&exception);

  im_image_info = Magick::CloneImageInfo((Magick::ImageInfo *) NULL);
  (void) strcpy(im_image_info -> filename, file_name.c_str());

  mg_log << TRACE << "Reading image by ImageMagick" << endlog;

  im_image = Magick::ReadImage(im_image_info, &exception);

  if (exception.severity != Magick::UndefinedException)
    if (exception.severity >= Magick::ErrorException)
      throw ExImageTransferError("ImageMagickReadError", (std::string)exception.reason + ": " + exception.description);
    else
      mg_log << WARN << (std::string)exception.reason + ": " + exception.description << endlog;

  if (im_image == (Magick::Image *) NULL)
    throw ExImageTransferError("ImageMagickReadError", "Cannot read image by ImageMagick - unspecified error");

  Magick::DestroyImageInfo(im_image_info);
  Magick::DestroyExceptionInfo(&exception);

  return im_image;
  }
//-----------------------------------------------------------------------------
void ImageTransfer::Write(const std::string& file_name, Magick::Image* im_image)
  {
  Magick::ExceptionInfo exception;
  
  Magick::ImageInfo *im_image_info;

  Magick::GetExceptionInfo(&exception);

  im_image_info = Magick::CloneImageInfo((Magick::ImageInfo *) NULL);

  if (m_format != "by-extention")
    {
    mg_log << TRACE << "Changing file format type to " << m_format << endlog;
    strcpy(im_image -> magick, m_format.c_str());
    }

  strcpy(im_image -> filename, file_name.c_str());
  im_image_info -> quality = m_quality;

  mg_log << TRACE << "Writing ImageMagick image" << endlog;

  Magick::WriteImage(im_image_info, im_image);

  if (im_image -> exception.severity != Magick::UndefinedException)
    if (im_image -> exception.severity >= Magick::ErrorException)
      throw ExImageTransferError("ImageMagickWriteError", (std::string)im_image -> exception.reason + ": " + im_image -> exception.description);
    else
      mg_log << WARN << (std::string)im_image -> exception.reason + ": " + im_image -> exception.description << endlog;

  mg_log << TRACE << "Destroying temporary info" << endlog;

  Magick::DestroyImageInfo(im_image_info);
  Magick::DestroyExceptionInfo(&exception);
  }
//-----------------------------------------------------------------------------
Magick::Image* ImageTransfer::ImageToIMImage(const Image& image, bool fl_trim)
  {
  Magick::ExceptionInfo exception;
  Magick::GetExceptionInfo(&exception);

  mg_log << TRACE << "Creating ImageMagick image from given one" << endlog;

  ulong trim;
  
  if (fl_trim)
    trim = (int)image.GetTrimWidth() + m_add_trim;
  else
    trim = 0;

  Magick::Image* im_image = Magick::ConstituteImage(image.GetWidth() - 2*trim, image.GetHeight() - 2*trim,
    "RGB", Magick::ShortPixel, reinterpret_cast<const void*>(&image.GetContainer()[0]),
    &exception);

  if (exception.severity != Magick::UndefinedException)
    if (exception.severity >= Magick::ErrorException)
      throw ExImageTransferError("ImageMagickImageToIMError", (std::string)exception.reason + ": " + exception.description);
    else
      mg_log << WARN << (std::string)exception.reason + ": " + exception.description << endlog;

  if (im_image == (Magick::Image *) NULL)
    throw ExImageTransferError("ImageMagickImageToIMError", "Cannot construct ImageMagick image - unspecified error");

  if (fl_trim)
    {
    mg_log << TRACE << "Trimming ImageMagick image" << endlog;

    Magick::PixelPacket* pp = Magick::GetImagePixels(im_image, 0, 0, image.GetWidth() - 2*trim,
      image.GetHeight() - 2*trim);

    for (unsigned row = 0; row < image.GetHeight() - 2*trim; ++row)
      for (unsigned col = 0; col < image.GetWidth() - 2*trim; ++col, ++pp)
	{
	pp -> red = image.At(row + trim, col + trim)[0];
	pp -> green = image.At(row + trim, col + trim)[1];
	pp -> blue = image.At(row + trim, col + trim)[2];
	}
    }

  if (!image.GetProfile().empty())
    {
    mg_log << TRACE << "Copying embedded profile into ImageMagick image" << endlog;

    Magick::ProfileImage(im_image, "ICM", reinterpret_cast<const uchar*>(image.GetProfile().c_str()),
			image.GetProfile().length(), Magick::MagickTrue);
    }

  if (im_image -> exception.severity != Magick::UndefinedException)
    if (im_image -> exception.severity >= Magick::ErrorException)
      throw ExImageTransferError("ImageMagickWriteError", (std::string)im_image -> exception.reason + ": " + im_image -> exception.description);
    else
      mg_log << WARN << (std::string)im_image -> exception.reason + ": " + im_image -> exception.description << endlog;


  Magick::DestroyExceptionInfo(&exception);

  return im_image;
  }
//-----------------------------------------------------------------------------
void ImageTransfer::IMImageToImage(Image& image, Magick::Image* im_image)
  {
  Magick::ExceptionInfo exception;

  Magick::ImageInfo *im_image_info;

  Magick::GetExceptionInfo(&exception);

  im_image_info = Magick::CloneImageInfo((Magick::ImageInfo *) NULL);

  mg_log << TRACE << "Transferring image to internal representation" << endlog;

  // Let's transfer im_image to image
  image.Create(im_image -> columns, im_image -> rows, 0xFFFF, 0);

  Magick::DispatchImage (im_image, 0, 0, image.GetWidth(), image.GetHeight(),
    "RGB", Magick::ShortPixel, &image.GetContainer()[0], &exception);

  if (exception.severity != Magick::UndefinedException)
    if (exception.severity >= Magick::ErrorException)
      throw ExImageTransferError("ImageMagickIMtoImageError", (std::string)exception.reason + ": " + exception.description);
    else
      mg_log << WARN << (std::string)exception.reason + ": " + exception.description << endlog;

  Magick::DestroyImageInfo(im_image_info);
  Magick::DestroyExceptionInfo(&exception);
  }
//-----------------------------------------------------------------------------
#endif // USE_IMAGE_MAGICK
