/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * ImageUtilities - misc image processing algorithms that are used as parts
 * of more complicated ones or just as an items of workflow.
******************************************************************************/

#include "stdafx.h"
#include "ImageUtilities.h"
#include "WBCalculator.h"

#include "RawImageInfo.h"
#include "Image.h"
#include <lcms.h>

#include <boost/tokenizer.hpp>
//-----------------------------------------------------------------------------
void ImageUtilities::Process(Image& image, const RawImageInfo& image_info, const std::string& method)
  {
  if (method == "ApplyGamma()")
    ApplyGamma(image, m_gamma);
  else if (method == "CorrectExposure()")
    CorrectExposure(image, m_expo_correction);
  else
    {
    typedef boost::tokenizer<boost::char_separator<char> > Tokenizer;
    boost::char_separator<char> sep("(), \t");
    Tokenizer tokens(method, sep);

    if (*tokens.begin() == "EmbedColorProfile")
      EmbedColorProfile(image, *++tokens.begin());
    else if (*tokens.begin() == "ApplyColorProfile")
      ApplyColorProfile(image, *++tokens.begin(), *++++tokens.begin());
    else if (*tokens.begin() == "ApplyWB")
      {
      if (++tokens.begin() == tokens.end())
	ApplyWB(image, image_info.GetRedWB(), image_info.GetGreenWB(), image_info.GetBlueWB());
      else
	ApplyWB(image, boost::lexical_cast<float>(*++tokens.begin()),
		boost::lexical_cast<float>(*++++tokens.begin()),
		boost::lexical_cast<float>(*++++++tokens.begin())); // Yes, I'm pervertor, I know ;-)
      }
    else if (*tokens.begin() == "SubtractNoise")
	  SubtractNoise(image, boost::lexical_cast<float>(*++tokens.begin()),
		            boost::lexical_cast<float>(*++++tokens.begin()),
		            boost::lexical_cast<float>(*++++++tokens.begin()));
    else if (*tokens.begin() == "ApplyWBColorTemp")
      ApplyWBColorTemp(image, boost::lexical_cast<float>(*++tokens.begin()),
		       boost::lexical_cast<float>(*++++tokens.begin()),
               *++++++tokens.begin());

    else throw ExImageProcessorError("UnknMethod", "Attempting to call unexisting processing method ImageUtilities::" + method);
    }
  }
//-----------------------------------------------------------------------------
ImageUtilities::ImageUtilities(): IParametrizable("ImageUtilities")
  {
  m_gamma = 1.0;
  m_expo_correction = 0.0;
  }
//-----------------------------------------------------------------------------
ImageUtilities& ImageUtilities::GetImageProcessor()
  {
  static ImageUtilities utilities;
  return utilities;
  }
//-----------------------------------------------------------------------------
void ImageUtilities::GetParameters(Params& o_params) const
  {
  o_params["ImageUtilities_Gamma"] = boost::lexical_cast<std::string>(m_gamma);
  o_params["ImageUtilities_ExpoCorrection"] = boost::lexical_cast<std::string>(m_expo_correction);
  }
//-----------------------------------------------------------------------------
void ImageUtilities::GetDescriptions(Params& o_params) const
  {
  o_params["ImageUtilities::ApplyGamma()"] = "Image gamma correction";
  o_params["ImageUtilities::CorrectExposure()"] = "Exposure correction";
  o_params["ImageUtilities_Gamma"] = "The gamma to apply to output image";
  o_params["ImageUtilities_ExpoCorrection"] = "The exposure correction to apply to input image";
  }
//-----------------------------------------------------------------------------
void ImageUtilities::SetParameters(Params& i_params)
  {
  if (i_params.exist("ImageUtilities_Gamma"))
    m_gamma = boost::lexical_cast<float>(i_params.pop("ImageUtilities_Gamma"));
  if (i_params.exist("ImageUtilities_ExpoCorrection"))
    m_expo_correction = boost::lexical_cast<float>(i_params.pop("ImageUtilities_ExpoCorrection"));
  }
//-----------------------------------------------------------------------------
void ImageUtilities::SubtractNoise(Image& image, ushort red, ushort green, ushort blue)
  {
  mg_log << INFO << "Subtracting thermal noise" << endlog;
  mg_log << TRACE << "Red noise: " << red
                  << ", green noise: " << green
                  << ", blue noise: "  << blue << endlog;
    
  ushort noise[3] = {red, green, blue};

  for (unsigned row = 0; row < image.GetHeight(); ++row)
    for (unsigned col = 0; col < image.GetWidth(); ++col)
      for (uchar i = 0; i < 3; ++i)
	if (image.At(row, col)[i] < noise[i])
	  image.At(row, col)[i] = 0;
	else image.At(row, col)[i] -= noise[i];

  mg_log << TRACE << "Finished" << endlog;
  }
//-----------------------------------------------------------------------------
void ImageUtilities::ApplyWB(Image& image, float red, float green, float blue)
  {
  mg_log << INFO << "Applying White Balance: " <<
    "Red=" << red << ", Green=" << green << ", Blue=" << blue << endlog;

  for (unsigned row = 0; row < image.GetHeight(); ++row)
    for (unsigned col = 0; col < image.GetWidth(); ++col)
      {
      image.At(row, col)[0] = std::min((int)(image.At(row, col)[0] * red), (int)image.GetClampVal());

      if (green != 1.0)
	image.At(row, col)[1] = std::min((int)(image.At(row, col)[1] * green), (int)image.GetClampVal());

      image.At(row, col)[2] = std::min((int)(image.At(row, col)[2] * blue), (int)image.GetClampVal());
      }
  }
//-----------------------------------------------------------------------------
WhiteBalance ImageUtilities::GetWhiteBalanceFromColorTemp(float temp, float shift, const std::string& profile)
  {
  mg_log << INFO << "Calculating White Balance from color temperature " <<
    temp << " and shift " << shift << endlog;

  PlanckianCalculator::PlanckianPoint pnt = PlanckianCalculator::ProjectOnPlanckian(temp);
  PlanckianCalculator::PlanckianPoint pnt_et = PlanckianCalculator::ProjectOnPlanckian(2850);
	//PlanckianCalculator::PlanckianPoint pnt_et = PlanckianCalculator::ProjectOnPlanckian(6500);

  Pixel_xyz pix_xyz, pix_et_xyz;
  pix_xyz[0] = pnt.m_chroma.m_r/pnt.m_chroma.m_g;
  pix_xyz[1] = 1.0;
  pix_xyz[2] = (1 - pnt.m_chroma.m_r - pnt.m_chroma.m_g)/pnt.m_chroma.m_g;

  pix_et_xyz[0] = pnt_et.m_chroma.m_r/pnt_et.m_chroma.m_g;
  pix_et_xyz[1] = 1.0;
  pix_et_xyz[2] = (1 - pnt_et.m_chroma.m_r - pnt_et.m_chroma.m_g)/pnt_et.m_chroma.m_g;

  PixelFloat pnt_rgb, pnt_et_rgb;

  mg_log << INFO << "Using " << profile << " for info about primaries" << endlog;

  bool fl_converted = false;

  cmsHPROFILE h_profile = cmsOpenProfileFromFile(profile.c_str(), "r");

  if (cmsIsTag(h_profile, icSigRedColorantTag) &&
      cmsIsTag(h_profile, icSigGreenColorantTag) &&
	  cmsIsTag(h_profile, icSigBlueColorantTag))
    {
    MAT3 matrix;

    if (cmsReadICCMatrixRGB2XYZ(&matrix, h_profile))
      {
  	  cmsCIEXYZ media_white;
      cmsTakeMediaWhitePoint(&media_white, h_profile);
	  cmsCIExyY media_wpnt;
	  cmsXYZ2xyY(&media_wpnt, &media_white);

      // I don't think we need this:
      // cmsAdaptMatrixFromD50(&matrix, &media_wpnt);

	  cmsCIEXYZ Red, Green, Blue;
      cmsCIExyYTRIPLE primaries;

      Red.X = matrix.v[0].n[0];
      Red.Y = matrix.v[1].n[0];
      Red.Z = matrix.v[2].n[0];

      cmsXYZ2xyY(&primaries.Red, &Red);

      Green.X = matrix.v[0].n[1];
      Green.Y = matrix.v[1].n[1];
      Green.Z = matrix.v[2].n[1];

      cmsXYZ2xyY(&primaries.Green, &Green);

      Blue.X = matrix.v[0].n[2];
      Blue.Y = matrix.v[1].n[2];
      Blue.Z = matrix.v[2].n[2];

      cmsXYZ2xyY(&primaries.Blue, &Blue);

      MAT3 prim, trans;

      VEC3init(&prim.v[0], primaries.Red.x, primaries.Red.y, 1 - primaries.Red.x - primaries.Red.y);
      VEC3init(&prim.v[1], primaries.Green.x, primaries.Green.y, 1 - primaries.Green.x - primaries.Green.y);
      VEC3init(&prim.v[2], primaries.Blue.x, primaries.Blue.y, 1 - primaries.Blue.x - primaries.Blue.y);

      mg_log << DEBUG << "Primaries matrix (xyY):" << endlog;
      mg_log << DEBUG << prim.v[0].n[0] << ", " << prim.v[0].n[1] << ", " << prim.v[0].n[2] << endlog;
      mg_log << DEBUG << prim.v[1].n[0] << ", " << prim.v[1].n[1] << ", " << prim.v[1].n[2] << endlog;
      mg_log << DEBUG << prim.v[2].n[0] << ", " << prim.v[2].n[1] << ", " << prim.v[2].n[2] << endlog;

      VEC3init(&prim.v[0], Red.X, Red.Y, Red.Z);
      VEC3init(&prim.v[1], Green.X, Green.Y, Green.Z);
      VEC3init(&prim.v[2], Blue.X, Blue.Y, Blue.Z);

      mg_log << DEBUG << "Primaries matrix (XYZ):" << endlog;
      mg_log << DEBUG << prim.v[0].n[0] << ", " << prim.v[0].n[1] << ", " << prim.v[0].n[2] << endlog;
      mg_log << DEBUG << prim.v[1].n[0] << ", " << prim.v[1].n[1] << ", " << prim.v[1].n[2] << endlog;
      mg_log << DEBUG << prim.v[2].n[0] << ", " << prim.v[2].n[1] << ", " << prim.v[2].n[2] << endlog;

      if (MAT3inverse(&prim, &trans) != -1)
        {
        mg_log << DEBUG << "Inverse primaries matrix:" << endlog;
        mg_log << DEBUG << trans.v[0].n[0] << ", " << trans.v[0].n[1] << ", " << trans.v[0].n[2] << endlog;
        mg_log << DEBUG << trans.v[1].n[0] << ", " << trans.v[1].n[1] << ", " << trans.v[1].n[2] << endlog;
        mg_log << DEBUG << trans.v[2].n[0] << ", " << trans.v[2].n[1] << ", " << trans.v[2].n[2] << endlog;

        for (uchar i = 0; i < 3; ++i)
          {
          pnt_rgb[i] = trans.v[0].n[i]*pix_xyz[0] +
                       trans.v[1].n[i]*pix_xyz[1] +
                       trans.v[2].n[i]*pix_xyz[2];

          pnt_et_rgb[i] = trans.v[0].n[i]*pix_et_xyz[0] +
                          trans.v[1].n[i]*pix_et_xyz[1] +
                          trans.v[2].n[i]*pix_et_xyz[2];
          }

        fl_converted = true;
        }
      }
	}

  cmsCloseProfile(h_profile);

  if (!fl_converted)
    {
    mg_log << CRIT << "Cannot convert temperature chromacity into camera RGB: possibly problems with profile?" << endlog;
    throw ExImageProcessorError("ErrorWBColorTemp", "Cannot convert temperature chromacity into camera RGB: possibly problems with profile?");
    }

  mg_log << DEBUG << "Converted illuminant: " << pnt_rgb[0] << ", " << pnt_rgb[1] << ", " << pnt_rgb[2] << endlog;
  mg_log << DEBUG << "Converted etalon illuminant: " << pnt_et_rgb[0] << ", " << pnt_et_rgb[1] << ", " << pnt_et_rgb[2] << endlog;

  WhiteBalance wb;
  wb.m_red = pnt_rgb[1]/pnt_rgb[0]*pnt_et_rgb[0]/pnt_et_rgb[1];
  wb.m_green = 1.0;
  wb.m_blue = pnt_rgb[1]/pnt_rgb[2]* pnt_et_rgb[2]/pnt_et_rgb[1];

  /*
  5500?
  red *= 1.9717;
  blue *= 1.2378;
  */

  /*
  6000:
  red *= 2.5;
  blue *= 1.25;
  */

  /*
  2850
  red *= 1.16667;
  blue *= 2.3333;
  */
  wb.m_red *= 1.17155987496866;
  wb.m_blue *= 2.28818644364185;

  return wb;
  }
//-----------------------------------------------------------------------------
void ImageUtilities::ApplyWBColorTemp(Image& image, float temp, float shift, const std::string& profile)
  {
  mg_log << INFO << "Applying White Balance using color temperature " <<
    temp << " and shift " << shift << endlog;

  WhiteBalance wb = GetWhiteBalanceFromColorTemp(temp, shift, profile);

  ApplyWB(image, wb.m_red, wb.m_green, wb.m_blue);
  }
//-----------------------------------------------------------------------------
void ImageUtilities::CorrectExposure(Image& image, float coef)
  {
  if (coef == 0.0)
    {
    mg_log << INFO << "Exposure correction was set to 0.0 - no correction is performed" << endlog;
    return;
    }

  mg_log << INFO << "Correcting Exposure: " << coef << endlog;

  for (unsigned row = 0; row < image.GetHeight(); ++row)
    for (unsigned col = 0; col < image.GetWidth(); ++col)
      for (uchar ind = 0; ind < 3; ++ind)
	image.At(row, col)[ind] = std::min((int)(image.At(row, col)[ind] * coef), (int)image.GetClampVal());
  }
//-----------------------------------------------------------------------------
void ImageUtilities::ConvertRGB2RGGB(Image& image)
  {
  mg_log << INFO << "Converting RGB to chroma" << endlog;
  for (unsigned row = 0; row < image.GetHeight(); ++row)
    for (unsigned col = 0; col < image.GetWidth(); ++col)
      {
      image.At(row, col)[0] = ((int)image.GetClampVal() + (int)image.At(row, col)[1] - (int)image.At(row, col)[0]) / 2;
      image.At(row, col)[2] = ((int)image.GetClampVal() + (int)image.At(row, col)[1] - (int)image.At(row, col)[2]) / 2;
      }
  }
//-----------------------------------------------------------------------------
void ImageUtilities::ConvertRGGB2RGB(Image& image)
  {
  mg_log << INFO << "Converting chroma to RGB" << endlog;
  for (unsigned row = 0; row < image.GetHeight(); ++row)
    for (unsigned col = 0; col < image.GetWidth(); ++col)
      {
      int val = (int)image.GetClampVal() + (int)image.At(row, col)[1] - 2*(int)image.At(row, col)[0];

      if (val > 0)
	image.At(row, col)[0] = std::min(val, (int)image.GetClampVal());
      else
	image.At(row, col)[0] = 0;

      val = (int)image.GetClampVal() + (int)image.At(row, col)[1] - 2*(int)image.At(row, col)[2];

      if (val > 0)
	image.At(row, col)[2] = std::min(val, (int)image.GetClampVal());
      else
	image.At(row, col)[2] = 0;
      }
  }
//-----------------------------------------------------------------------------
void ImageUtilities::ApplyGamma(Image& image, float gamma)
  {
  if (gamma == 1.0)
    {
    mg_log << INFO << "Gamma was set to 1.0 - no gamma correction is performed" << endlog;
    return;
    }

  mg_log << INFO << "Applying gamma " << gamma << endlog;

  std::vector<float> vals;
  vals.resize(image.GetClampVal() + 1);

  vals[0] = 1.0;

  for (unsigned i = 1; i <= image.GetClampVal(); ++i)
    vals[i] = pow(i/(float)image.GetClampVal(), 1.0f/gamma - 1.0f);

  for (unsigned row = 0; row < image.GetHeight(); ++row)
    for (unsigned col = 0; col < image.GetWidth(); ++col)
      {
      for (uchar ind = 0; ind < 3; ++ind)
	{
	ushort i = (int)image.At(row, col)[ind];

	if (i > image.GetClampVal())
	  i = image.GetClampVal();

	int val = (int)(image.At(row, col)[ind] * vals[i]);

	if (val > image.GetClampVal())
	  val = image.GetClampVal();

	image.At(row, col)[ind] = val;
	}
      }
  }
//-----------------------------------------------------------------------------
void ImageUtilities::BuildHistogram(std::vector<PixelLong>& histogram, ulong size, const Image& image, bool fl_count_zeroes)
  {
  mg_log << INFO << "Building histogram" << endlog;
  ulong coef;

  if (!(image.GetClampVal()%size))
    coef = image.GetClampVal()/size;
  else
    coef = image.GetClampVal()/size + 1;

  histogram.resize(size);

  for (unsigned i = 0; i < size; ++i)
    {
    histogram[i][0] = 0;
    histogram[i][1] = 0;
    histogram[i][2] = 0;
    }

  for (unsigned row = 0; row < image.GetHeight(); ++row)
    for (unsigned col = 0; col < image.GetWidth(); ++col)
      for (uchar ind = 0; ind < 3; ++ind)
	{
	if (!fl_count_zeroes && !image.At(row, col)[ind])
	  continue;

	++histogram[image.At(row, col)[ind]/coef][ind];
	}
  }
//-----------------------------------------------------------------------------
void ImageUtilities::EmbedColorProfile(Image& image, const std::string& name)
  {
  mg_log << INFO << "Embedding profile " << name << " into image" << endlog;

  if (!image.GetProfile().empty())
    mg_log << WARN << "Embedding of this profile will destroy old profile in the image" << endlog;

  std::ifstream prof(name.c_str(), std::ios::in | std::ios::binary);

  if (!prof.is_open())
    {
    mg_log << ERR << "Cannot open the specified profile - skipping embedding" << endlog;
    return;
    }

  std::string profile;

  for (char c = prof.get(); !prof.eof(); c = prof.get())
    profile.append(1, c);
  prof.close();

  mg_log << DEBUG << "Profile size is " << (unsigned)profile.length() << endlog;

  image.SetProfile(profile);
  }
//-----------------------------------------------------------------------------
void ImageUtilities::ApplyColorProfile(Image& image, const std::string& in_profile,
				       const std::string& out_profile)
  {
  mg_log << INFO << "Applying color transformation from " << in_profile << " to " << out_profile << endlog;

  cmsHPROFILE hInProfile, hOutProfile;
  cmsHTRANSFORM hTransform;

  hInProfile  = cmsOpenProfileFromFile(in_profile.c_str(), "r");
  hOutProfile = cmsOpenProfileFromFile(out_profile.c_str(), "r");

  hTransform = cmsCreateTransform(hInProfile,
				  TYPE_RGB_16,
				  hOutProfile,
				  TYPE_RGB_16,
				  INTENT_PERCEPTUAL, 0);

  mg_log << DEBUG << "Finished transformation creation" << endlog;

  image.StartBuffer(0, 1);

  for (unsigned row = 0; row < image.GetHeight(); ++row)
    {
    cmsDoTransform(hTransform, &image.At(row, 0),
		   &image.AtBuffer(row, 0),
		   image.GetWidth());

    image.FlushBufferRow();
    }

  image.StopBuffer();

  mg_log << DEBUG << "Finished, destroying temporary info" << endlog;

  cmsDeleteTransform(hTransform);
  cmsCloseProfile(hInProfile);
  cmsCloseProfile(hOutProfile);

  //if (!image.GetProfile().empty())
  //  {
  image.SetProfile("");
  EmbedColorProfile(image, out_profile);
  //  }
  }
//-----------------------------------------------------------------------------
