/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * Processor - main class that dispatches image processing calls.
******************************************************************************/

#include "stdafx.h"

#include "Processor.h"

#include "Image.h"
#include "RawFileReader.h"
#include "RawImageInfo.h"
#include "RawInterpolator.h"
#include "ParamsDispatcher.h"

#include "CRWImageInfo.h"
#include "CRWFileReader.h"

#include "ACPInterpolator.h"
#include "BilinearInterpolator.h"
#include "CornerInterpolator.h"
#include "HueGradInterpolator.h"
#include "MedianInterpolator.h"
#include "MultiGradientInterpolator.h"
#include "WeightedEdgeInterpolator.h"

#include "WBGreyWorldCalculator.h"
#include "WBScaleByMaxCalculator.h"
#include "WBDichromaticCalculator.h"

#include "ImageTransfer.h"

#include "ImageSimpleFilters.h"
#include "ImageUtilities.h"
#include "WeightedEdgeImageProcessor.h"
#include "ChromaDiffFlowImageProcessor.h"
#include "TVImageProcessor.h"

#include <boost/tokenizer.hpp>

//-----------------------------------------------------------------------------
//#define WRITE_RAW_IMAGE
#undef WRITE_RAW_IMAGE
//-----------------------------------------------------------------------------
Processor::Processor(): IParametrizable("Processor"), mp_image_info(NULL)
  {
  mg_log << TRACE << "Constructing engine" << endlog;

  m_interpolators["ACPInterpolator"] = &ACPInterpolator::GetInterpolator();
  m_interpolators["BilinearInterpolator"] = &BilinearInterpolator::GetInterpolator();
  m_interpolators["CornerInterpolator"] = &CornerInterpolator::GetInterpolator();
  m_interpolators["HueGradInterpolator"] = &HueGradInterpolator::GetInterpolator();
  m_interpolators["MedianInterpolator"] = &MedianInterpolator::GetInterpolator();
  m_interpolators["MultiGradientInterpolator"] = &MultiGradientInterpolator::GetInterpolator();
  m_interpolators["WeightedEdgeInterpolator"] = &WeightedEdgeInterpolator::GetInterpolator();

  m_readers["CRWFileReader"] = &CRWFileReader::GetRawReader();

  m_wb_calculators["WBGreyWorldCalculator"] = &WBGreyWorldCalculator::GetWBCalculator();
  m_wb_calculators["WBScaleByMaxCalculator"] = &WBScaleByMaxCalculator::GetWBCalculator();
  m_wb_calculators["WBDichromaticCalculator"] = &WBDichromaticCalculator::GetWBCalculator();

  m_image_processors["ImageSimpleFilters"] = &ImageSimpleFilters::GetImageProcessor();
  m_image_processors["ImageUtilities"] = &ImageUtilities::GetImageProcessor();
  m_image_processors["WeightedEdgeImageProcessor"] = &WeightedEdgeImageProcessor::GetImageProcessor();
  m_image_processors["ChromaDiffFlowImageProcessor"] = &ChromaDiffFlowImageProcessor::GetImageProcessor();
  m_image_processors["TVImageProcessor"] = &TVImageProcessor::GetImageProcessor();

  mg_log << TRACE << "Engine construction completed" << endlog;

//  DumpSupportedFileFormats();
//  DumpDefaultParameters();
//  DumpParametersDescriptions();
  }
//-----------------------------------------------------------------------------
Processor::~Processor()
  {
  delete mp_image_info;
  }
//-----------------------------------------------------------------------------
Processor& Processor::GetProcessor()
  {
  static Processor proc;
  return proc;
  }
//-----------------------------------------------------------------------------
void Processor::GetParameters(Params& o_params) const
  {
  }
//-----------------------------------------------------------------------------
void Processor::GetDescriptions(Params& o_params) const
  {
  o_params["Processor::DumpDefaultParams()"] = "Dumps default parameters";
  o_params["Processor::DumpParamsDescription()"] = "Dumps parameters description";
  o_params["Processor::DumpSupportedFormats()"] = "Dumps supported image formats";
  }
//-----------------------------------------------------------------------------
void Processor::SetParameters(Params& i_params)
  {
  }
//-----------------------------------------------------------------------------
void Processor::DumpSupportedFileFormats()
  {
  mg_log << TRACE << "Supported file formats:" << std::endl;

  std::list<std::string> formats;
  ImageTransfer::GetTransfer().GetSupportedFormats(formats);

  for (std::list<std::string>::const_iterator it = formats.begin(); it != formats.end(); ++it)
    mg_log << *it << std::endl;

  mg_log << "" << endlog;
  }
//-----------------------------------------------------------------------------
void Processor::DumpDefaultParameters()
  {
  ParamsDispatcher::ParamsList list;
  ParamsDispatcher::GetDispatcher().GetParameters(list);

  for (ParamsDispatcher::ParamsList::const_iterator it = list.begin(); it != list.end(); ++it)
    {
    mg_log << INFO << "Module " << it -> first << ":" << endlog;

    for (Params::const_iterator it1 = it -> second.begin(); it1 != it -> second.end(); ++it1)
      mg_log << INFO << "    " << it1 -> first << " = " << it1 -> second << endlog;
    }
  }
//-----------------------------------------------------------------------------
void Processor::DumpParametersDescriptions()
  {
  ParamsDispatcher::ParamsList list;
  ParamsDispatcher::GetDispatcher().GetDescriptions(list);

  for (ParamsDispatcher::ParamsList::const_iterator it = list.begin(); it != list.end(); ++it)
    {
    mg_log << INFO << "Module " << it -> first << ":" << endlog;

    for (Params::const_iterator it1 = it -> second.begin(); it1 != it -> second.end(); ++it1)
      mg_log << INFO << "    " << it1 -> first << " = " << it1 -> second << endlog;
    }
  }
//-----------------------------------------------------------------------------
void Processor::LoadImage(const std::string& file_name, bool fl_parse_only,
                          bool fl_small_icon_mode)
  {
  try
    {
    bool fl_processed = false;

    if (mp_image_info)
      {
      delete mp_image_info;
      mp_image_info = NULL;
      }

    m_thumbnails.clear();

    for (std::map<std::string, RawFileReader*>::iterator it = m_readers.begin(); it != m_readers.end(); ++it)
      try
        {
        RawFileReader* p_reader = it -> second;
        p_reader -> LoadRaw(file_name, m_image, fl_parse_only,
                            fl_small_icon_mode);

        mp_image_info = p_reader -> RetrieveImageInfo();

        if (fl_parse_only && fl_small_icon_mode
            && mp_image_info -> GetThumbnails().empty())
          {
          delete mp_image_info;
          mp_image_info = NULL;

          p_reader -> LoadRaw(file_name, m_image, fl_parse_only,
                              false);
          mp_image_info = p_reader -> RetrieveImageInfo();
          }

        fl_processed = true;

        if (!fl_parse_only)
          {
          ImageUtilities::GetImageProcessor().SubtractNoise(m_image, mp_image_info -> GetNoise(0),
                                                mp_image_info -> GetNoise(1), mp_image_info -> GetNoise(2));
/*
          ImageUtilities::GetImageProcessor().ApplyColorProfile(m_image, "s45-no-wb.icm", "sRGB.icm");

          WBCalculator::WhiteBalance wb;
          wb = m_wb_calculators["WBGreyWorldCalculator"] -> CalculateWB(m_image, *mp_image_info);

          p_reader -> LoadRaw(file_name, m_image, fl_parse_only,
                              fl_small_icon_mode);

          ImageUtilities::GetImageProcessor().ApplyWB(m_image, wb.m_red, wb.m_green, wb.m_blue);
*/
/*
ImageUtilities::GetImageProcessor().ApplyWB(m_image,
            mp_image_info ->GetRedWB(),
            mp_image_info ->GetGreenWB(),
            mp_image_info ->GetBlueWB());
*/

#ifdef WRITE_RAW_IMAGE
          ImageTransfer::GetTransfer().Write("raw.tif", m_image);
#endif
          }
        break;
        }
      catch (RawFileReader::ExRawFileError err)
        {
        if (err.GetID() != "UnsupportedCamera")
          throw;
        }

    if (!fl_processed)
      throw ExProcessorError("UnsupportedCamera", "All RAW readers refused to read this RAW file");

#ifdef USE_IMAGE_MAGICK
    if (!mp_image_info -> GetThumbnails().empty())
      {
      mg_log << TRACE << "Extracting JPEG thumbnails" << endlog;

      const char* thmb;
      ulong len;

      for (ulong i = 0; i < mp_image_info -> GetThumbnails().size(); ++i)
        {
        thmb = &mp_image_info -> GetThumbnails()[i][0];
        len = (ulong)mp_image_info -> GetThumbnails()[i].size();

        Magick::ExceptionInfo exception;
        Magick::GetExceptionInfo(&exception);

        Magick::ImageInfo *im_image_info = Magick::CloneImageInfo((Magick::ImageInfo *) NULL);

        strcpy(im_image_info -> magick, "JPEG");

        Magick::Image *im_image = Magick::BlobToImage(im_image_info, thmb, len, &exception);

        if (exception.severity != Magick::UndefinedException)
          mg_log << WARN << (std::string)exception.reason + ": " + exception.description << endlog;

        if (exception.severity < Magick::ErrorException &&
            im_image != (Magick::Image *) NULL)
          {
          Magick::Image* im_scaled_image = im_image;

          // Should we rescale image?
          if (fl_small_icon_mode && std::max(im_image -> columns, im_image -> rows) != 160)
            {
            if (im_image -> rows > im_image -> columns)
              im_scaled_image = Magick::ThumbnailImage(im_image, 160.0/im_image -> rows*im_image -> columns, 160, &exception);
            else
              im_scaled_image = Magick::ThumbnailImage(im_image, 160, 160.0/im_image -> columns*im_image -> rows, &exception);

            if (exception.severity != Magick::UndefinedException)
              mg_log << WARN << (std::string)exception.reason + ": " + exception.description << endlog;

            if (exception.severity >= Magick::ErrorException)
              im_scaled_image = (Magick::Image *) NULL;
            }

          Magick::Image* im_rotated_image = im_scaled_image;

          if (im_scaled_image)
            {
            if (mp_image_info -> GetRotationAngle() != 0)
              {
              im_rotated_image = Magick::RotateImage(im_scaled_image,
                 mp_image_info -> GetRotationAngle(), &exception);

              if (exception.severity != Magick::UndefinedException)
                mg_log << WARN << (std::string)exception.reason + ": " + exception.description << endlog;

              if (exception.severity >= Magick::ErrorException)
                im_rotated_image = (Magick::Image *) NULL;
              }

            if (im_rotated_image)
              {
              m_thumbnails.push_back(Image());
              ImageTransfer::GetTransfer().IMImageToImage(m_thumbnails.back(), im_rotated_image);
              }
            }

          if (im_rotated_image && im_rotated_image != im_scaled_image)
            Magick::DestroyImage(im_rotated_image);

          if (im_scaled_image && im_scaled_image != im_image)
            Magick::DestroyImage(im_scaled_image);

          Magick::DestroyImage(im_image);
          }

        Magick::DestroyImageInfo(im_image_info);
        Magick::DestroyExceptionInfo(&exception);
        }
      }
#endif// USE_IMAGE_MAGICK
    }
  catch(RawFileReader::ExRawFileError err)
    {
    if (fl_parse_only || err.GetID() != "IncorrectRAWFormat")
      throw;

    mg_log << WARN << "File " << file_name << " is not in supported RAW format" << endlog;
    mg_log << WARN << "Trying to read it as image file" << endlog;

    ImageTransfer::GetTransfer().Read(file_name, m_image);

    mp_image_info = new RawImageInfo;

    mp_image_info -> m_noise[0] = 0;
    mp_image_info -> m_noise[1] = 0;
    mp_image_info -> m_noise[2] = 0;
    }
  }
//-----------------------------------------------------------------------------
void Processor::Process(const std::string& str)
  {
  typedef boost::tokenizer<boost::char_separator<char> > Tokenizer;
  boost::char_separator<char> sep("; \t"), sep_eq("= \t"), sep_in(":");

  Tokenizer tokens(str, sep);

  for (Tokenizer::iterator it = tokens.begin(); it != tokens.end(); ++it)
    {
    mg_log << DEBUG << "Parsed token: " << *it << endlog;

    if (it -> find('=') != std::string::npos)
      {
      mg_log << DEBUG << "Parameter assignment detected" << endlog;

      Tokenizer tokens_eq(*it, sep_eq);

      std::string key = *tokens_eq.begin(), val = *++tokens_eq.begin();

      mg_log << DEBUG << "Parsed 1st item: " << key << endlog;
      mg_log << DEBUG << "Parsed 2nd item: " << val << endlog;

      Params params;
      params[key] = val;

      ParamsDispatcher::GetDispatcher().SetParameters(params);
      }
    else
      {
      mg_log << DEBUG << "Execution request detected" << endlog;

      Tokenizer tokens_in(*it, sep_in);

      if (tokens_in.begin() == tokens_in.end() ||
          ++tokens_in.begin() == tokens_in.end())
        throw ExProcessorError("SyntaxError", "Cannot parse user request \"" + *it + "\"");

      std::string key = *tokens_in.begin(), val = *++tokens_in.begin();

      mg_log << DEBUG << "Parsed 1st item: " << key << endlog;
      mg_log << DEBUG << "Parsed 2nd item: " << val << endlog;

      if (m_image_processors.find(key) != m_image_processors.end())
        m_image_processors[key] -> Process(m_image, *mp_image_info, val);
      else if (m_interpolators.find(key) != m_interpolators.end())
        m_interpolators[key] -> Interpolate(m_image, *mp_image_info);
      else if (key == "Processor")
        {
        if (val == "Load()")
          LoadImage(m_file_to_load);
        else if (val == "DumpDefaultParams()")
          DumpDefaultParameters();
        else if (val == "DumpParamsDescription()")
          DumpParametersDescriptions();
        else if (val == "DumpSupportedFormats()")
          DumpSupportedFileFormats();
        else throw ExProcessorError("UnknMethod", "Attempting to call unexisting processing method Processor::" + val);
        }
      else throw ExProcessorError("UnknClass", "Attempting to call method of unexisting class " + key);
      }
    }

  //WBCalculator::WhiteBalance wb;
//  try
//    {
  //wb = m_wb_calculators["WBScaleByMaxCalculator"] -> CalculateWB(m_image, *mp_image_info);
    //wb = m_wb_calculators["WBDichromaticCalculator"] -> CalculateWB(m_image, *mp_image_info);
//    }
//  catch(WBCalculator::ExWBError err)
//    {
//    mg_log << ERR << err << endlog;
//    mg_log << ERR << "Falling back to GreyWorld calculator" << endlog;
  //m_image.ApplyWB(1.71, 1.0, 1.95);
  //wb = m_wb_calculators["WBGreyWorldCalculator"] -> CalculateWB(m_image, *mp_image_info);
//    }

  //ImageUtilities::GetImageProcessor().ApplyWB(m_image, wb.m_red, wb.m_green, wb.m_blue);
  }
//-----------------------------------------------------------------------------
