/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * RawFileReader - implementation of RAW reader: general code.
 * It is heavily based on converter by Dave Coffin <dcoffin@shore.net>
******************************************************************************/

#include "stdafx.h"

#include "RawFileReader.h"
#include "RawImageInfo.h"
#include "Image.h"
//-----------------------------------------------------------------------------
//#define CIFF_DEBUG_INFO
//-----------------------------------------------------------------------------
RawFileReader::RawFileReader(): IParametrizable("RawFileReader")
  {
  m_trim_width = 16;
  m_fl_leave_border = false;
  m_fl_use_standard_size = false;
  }
//-----------------------------------------------------------------------------
void RawFileReader::GetParameters(Params& o_params) const
  {
  o_params["RawReader_TrimWidth"] = boost::lexical_cast<std::string>(m_trim_width);
  o_params["RawReader_FlLeaveBorder"] = boost::lexical_cast<std::string>(m_fl_leave_border);
  o_params["RawReader_FlStandardSize"] = boost::lexical_cast<std::string>(m_fl_use_standard_size);
  }
//-----------------------------------------------------------------------------
void RawFileReader::GetDescriptions(Params& o_params) const
  {
  o_params["RawReader_TrimWidth"] = "Width of additional border that will be "
    "added around the image before processing - will not be present in final result (for internal use only)";
  o_params["RawReader_FlLeaveBorder"] = "If true - image will be extracted "
    "from the RAW file\"as is\", including black border around it (mainly for developers only)";
  o_params["RawReader_FlStandardSize"] = "If true - standard output image size will be used "
    "(that is used by native converter)";
  }
//-----------------------------------------------------------------------------
void RawFileReader::SetParameters(Params& i_params)
  {
  if (i_params.exist("RawReader_TrimWidth"))
    m_trim_width = boost::lexical_cast<ulong>(i_params.pop("RawReader_TrimWidth"));
  if (i_params.exist("RawReader_FlLeaveBorder"))
    m_fl_leave_border = boost::lexical_cast<bool>(i_params.pop("RawReader_FlLeaveBorder"));
  if (i_params.exist("RawReader_FlStandardSize"))
    m_fl_use_standard_size = boost::lexical_cast<bool>(i_params.pop("RawReader_FlStandardSize"));
  }
//-----------------------------------------------------------------------------
ushort RawFileReader::Get2()
  {
  uchar a, b;

  a = m_file.get();
  b = m_file.get();

  if (m_order == 0x4d4d)          // "MM" means big-endian
    return (a << 8) + b;
  else                          // "II" means little-endian
    return a + (b << 8);
  }
//-----------------------------------------------------------------------------
ulong RawFileReader::Get4()
  {
  uchar a, b, c, d;

  a = m_file.get();
  b = m_file.get();
  c = m_file.get();
  d = m_file.get();

  if (m_order == 0x4d4d)
    return (a << 24) + (b << 16) + (c << 8) + d;
  else
    return a + (b << 8) + (c << 16) + (d << 24);
  }
//-----------------------------------------------------------------------------
void RawFileReader::CleanUp()
  {
  if (m_file.is_open())
    m_file.close();

  if (&GetImageInfo())
    delete RetrieveImageInfo();
  }
//-----------------------------------------------------------------------------
void RawFileReader::LoadRaw(const std::string& file_name, Image& image, bool fl_parse_only,
                            bool fl_small_icon_mode)
  {
  m_fl_parse_only = fl_parse_only;
  m_fl_small_icon_mode = fl_small_icon_mode;

  mg_log << TRACE << "Opening file " << file_name << endlog;
 
  if (m_file.is_open())
    {
    mg_log << WARN << "In RawFileReader: file was not closed after previous reading; will close it anyway" << endlog;
    m_file.close();
    }

  m_file.open(file_name.c_str(), std::ios::in | std::ios::binary);

  if (!m_file.is_open())
    {
    CleanUp();
    throw ExRawFileError("CannotOpenFile", "Cannot open file " + file_name);
    }

  m_order = Get2();
  long hlen = Get4();
  char head[8];
  m_file.read(head, 8);

  if (!memcmp(head, "HEAPCCDR", 8))
    {
    m_file.seekg(0, std::ios::end);
    if (&GetImageInfo())
      {
      mg_log << WARN << "RawImageInfo is not a NULL; destroying it explicitly" << endlog;
      delete RetrieveImageInfo();
      }
    PrepareImageInfo();
    m_file_length = m_file.tellg();
    ParseCiff(hlen, m_file_length - hlen);
    }
  else
    {
    CleanUp();
    throw ExRawFileError("IncorrectRAWFormat", "This RAW file has incorrect or unsupported format");
    }

  if (!m_fl_parse_only)
    {
    Load(image);

    mg_log << TRACE << "Filling trim zone" << endlog;

    // Let's fill trim zone:
    for (unsigned i = 0; i < m_trim_width; ++i)
      for (unsigned col = m_trim_width; col < image.GetWidth() - m_trim_width; ++col)
        {
        image.At(i, col) = image.At(2*m_trim_width - i, col);
        image.At(image.GetHeight() - i - 1, col) =
          image.At(image.GetHeight() - 2*m_trim_width + i - 1, col);
        }

    for (unsigned i = 0; i < m_trim_width; ++i)
      for (unsigned row = 0; row < image.GetHeight(); ++row)
        {
        image.At(row, i) = image.At(row, 2*m_trim_width - i);
        image.At(row, image.GetWidth() - i - 1) =
          image.At(row, image.GetWidth() - 2*m_trim_width + i - 1);
        }

    image.SetState(Image::RAW_SAMPLES);
    }

  m_file.close();
  }
//-----------------------------------------------------------------------------
void RawFileReader::ParseCiff(ulong offset, ulong length)
  {
  mg_log << TRACE << "Starting to parse CIFF" << endlog;

  Seek(offset + length - 4, std::ios::beg);
  ulong tboff = Get4() + offset;
  Seek(tboff, std::ios::beg);
  ushort rec_num = Get2();

  for (ushort i = 0; i < rec_num; ++i)
    {
    ushort type = Get2();
    ulong len = Get4();
    ulong rel_offset = Get4();
    ulong abs_offset = rel_offset + offset;
    std::streampos save_pos = m_file.tellg();

#ifdef CIFF_DEBUG_INFO
    mg_log.setf(std::ios::hex, std::ios::basefield);
    mg_log << TRACE << "CIFF tag: type=0x" << type << ", len=0x" << len << ", reloffset=" << rel_offset << endlog;
    mg_log.setf(std::ios::dec, std::ios::basefield);

    char buf[101],
      digits[16] = {'0', '1', '2', '3', '4', '5', '6', '7', '8',
                    '9', 'A', 'B', 'C', 'D', 'E', 'F'};
      Seek(abs_offset, std::ios::beg);
      m_file.read(buf, 100); buf[100] = 0;
      m_file.clear();
      m_file.seekg(save_pos);

      std::string res;
      for (unsigned i = 0; i < 100; ++i)
        {
        res += digits[(buf[i]>>4)&0x0F];
        res += digits[buf[i]&0x0F];

        if (!buf[i])
          buf[i] = '~';

        res.append(" ");
        }

      mg_log << TRACE << "content(text)=" << buf << endlog;
      mg_log << TRACE << "content(hex)=" << res << endlog;
      
#endif

    switch (type)
      {
      case 0x080A: // Camera model
        {
        mg_log << TRACE << "Camera model CIFF tag detected" << endlog;
	      char buf[64];
        Seek(abs_offset, std::ios::beg);
	      m_file.read(buf, 64);

	      char* zero_1st = std::find(buf, buf + 63, 0);
	      char* zero_2nd = std::find(zero_1st + 1, buf + 63, 0);
	      char* zero_1st_new = zero_1st;

	      while (*zero_1st_new == ' ') --zero_1st_new;
	      while (*zero_2nd == ' ') --zero_2nd;

	      GetImageInfo().m_manufacturer_name.assign(buf, zero_1st_new);
	      GetImageInfo().m_model_name.assign(zero_1st + 1, zero_2nd);

        mg_log << TRACE << "Manufacturer: \""<< GetImageInfo().GetManufacturerName() << "\"" << endlog;
        mg_log << TRACE << "Camera: \""<< GetImageInfo().GetModelName() << "\"" << endlog;
        }
      break;

      case 0x0810: // Owner name
        {
        mg_log << TRACE << "Owner name CIFF tag detected" << endlog;
        char buf[64];
        Seek(abs_offset, std::ios::beg);
	      m_file.read(buf, 64);
        GetImageInfo().m_owner_name.assign(buf);
        mg_log << TRACE << "Owner: "<< GetImageInfo().GetOwnerName() << endlog;
        }
        break;

      case 0x80b: // Firmware version
        {
        mg_log << TRACE << "Firmware version CIFF tag detected" << endlog;
        char buf[32];
        Seek(abs_offset, std::ios::beg);
	      m_file.read(buf, 32);
        std::string version = buf;
        
        mg_log << TRACE << "Version: "<< version << endlog;
        }
        break;

      case 0x1810:
        mg_log << TRACE << "Image specification CIFF tag detected" << endlog;
        Seek(abs_offset, std::ios::beg);
        GetImageInfo().m_std_width = Get4();
	      GetImageInfo().m_std_height = Get4();
        *(ulong*)&GetImageInfo().m_aspect_ratio = Get4();
        GetImageInfo().m_rotation_angle = Get4();

        mg_log << TRACE << "Standard image width, height: " << GetImageInfo().m_std_width << ", " << GetImageInfo().m_std_height << endlog;
        mg_log << TRACE << "Aspect ratio: " << GetImageInfo().m_aspect_ratio << endlog;
        mg_log << TRACE << "Rotation angle: " << GetImageInfo().m_rotation_angle << endlog;
        break;

      case 0x1031: // Raw width and height
        mg_log << TRACE << "Sensor size CIFF tag detected" << endlog;
        Seek(abs_offset + 2, std::ios::beg);
	      GetImageInfo().m_width = Get2();
	      GetImageInfo().m_height = Get2();

        Get4(); // skip unknown info

        GetImageInfo().m_left_border = Get2();
        GetImageInfo().m_top_border = Get2();

        mg_log << TRACE << "Reported left sensor offset: " << GetImageInfo().m_left_border << endlog;
        mg_log << TRACE << "Reported top sensor offset: " << GetImageInfo().m_top_border << endlog;

        mg_log << TRACE << "Raw image width, height: " << GetImageInfo().m_width << ", " << GetImageInfo().m_height << endlog;
	      break;

      default:
        if ((type >> 8) == 0x28 || (type >> 8) == 0x30)
          {
          mg_log << TRACE << "Sub-CIFF tag detected" << endlog;
	        ParseCiff(abs_offset, len);
          }
	      else
	        ParseCiffTag(type, len, abs_offset);
      }

    m_file.seekg(save_pos);
    }
  }
//-----------------------------------------------------------------------------
