/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#ifndef WB_CALCULATOR_H
#define WB_CALCULATOR_H

#include "Image.h"
#include "WhiteBalance.h"
#include "IParametrizable.h"
//-----------------------------------------------------------------------------
class RawImageInfo;
//-----------------------------------------------------------------------------
class PlanckianCalculator
  {
  public:
    struct Chromacity
      {
      float m_r, m_g;
      Chromacity(float r, float g): m_r(r), m_g(g) {}
      };

    struct PlanckianPoint
      {
      Chromacity m_chroma;
      float m_t;
      PlanckianPoint(float r, float g, float t): m_chroma(r, g), m_t(t)
	{};
      };

  public:
    static PlanckianPoint ProjectOnPlanckian(const Chromacity& ch);
  
    static PlanckianPoint ProjectOnPlanckian(float temp);

  private:
    static PlanckianPoint m_planckian_points[];
  };
//-----------------------------------------------------------------------------
class WBCalculator: public virtual IParametrizable, public PlanckianCalculator
  {
  public:
    class ExWBError: public MyException
      { 
      public:
	ExWBError(const std::string& i_err_id, const std::string& msg): MyException("WBCalculator", i_err_id, msg) {}
      };

    class WBSegmentPixelNotifier
      {
      public:
	virtual void NewPixelInSegment(ulong segment_number, const Position& pos) = 0;

	virtual void NewSegment(ulong segment_number) = 0;
      };

  public:
    virtual WhiteBalance CalculateWB(const Image& image, const RawImageInfo& image_info) = 0;

    virtual void GetParameters(Params& o_params) const;

    virtual void GetDescriptions(Params& o_params) const;

    virtual void SetParameters(Params& i_params);

  protected:
    class Mask
      {
      public:
	Mask(unsigned width, unsigned height): m_mask(width*height/8 + 1),
	  m_width(width), m_height(height)
	  {
	  std::fill(m_mask.begin(), m_mask.end(), 0);
	  }

	bool Get(unsigned row, unsigned col) const
	  {
	  ulong ind = (ulong)m_width*row + col;
	  return (m_mask[ind/8] >> (ind % 8)) & 1;
	  }

	void Set(unsigned row, unsigned col, bool val)
	  {
	  ulong ind = (ulong)m_width*row + col;
	  if (val)
	    m_mask[ind/8] |= (1 << (ind % 8)); // Set bit
	  else
	    m_mask[ind/8] &= ~(uchar)(1 << (ind % 8)); // Clear bit
	  }

      private:
	std::vector<uchar> m_mask;
	unsigned m_width, m_height;
      };

    ulong m_zone_width;
    float m_zone_threshold_dev;
    ushort m_threshold_dark, m_threshold_clamp,
      m_threshold_clamp2, m_threshold_clamp3;
    bool m_fl_use_planckian;

  protected:
    WBCalculator();

    void SegmentImage(WBSegmentPixelNotifier* p_notifier,
		      const Image& image, const RawImageInfo& image_info,
		      ulong max_segments_num);

    bool IsDeviated(const Image& image, unsigned row,
		    unsigned col, const RawImageInfo& image_info,
		    float threshold = 0, float* p_deviation = 0);

    bool IsClamped(const Image& image, unsigned row,
		   unsigned col, const RawImageInfo& image_info);

    void GetTheBestSeed(const Mask& mask,
			Position& pos, const Image& image,
			const RawImageInfo& image_info);

    bool IsGoodForSegment(const Mask& mask, const Image& image,
			  const RawImageInfo& image_info,
			  const Position& start_pos,
			  const Position& pos_to_test);
  };
//-----------------------------------------------------------------------------
#endif // WB_CALCULATOR_H

