/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * WeightedEdgeImageProcessor - image processing using hue correction and diffusion
 * flow, based on description in ``Demosaicing: Image reconstruction from color
 * CCD samples'' by Ron Kimmel.
******************************************************************************/

#include "stdafx.h"

#include "WeightedEdgeImageProcessor.h"
#include "Image.h"
#include "RawImageInfo.h"
//-----------------------------------------------------------------------------
WeightedEdgeImageProcessor::WeightedEdgeImageProcessor(): IParametrizable("WeightedEdgeImageProcessor")
  {
  /*
  Old params:
  m_alpha = 0.6;
  m_amount = 0.1;
  m_radius = 1.0;
  m_smooth_coef = 2000000;
  m_hue_num_of_iter = 3;
  m_diff_num_of_iter = 3;
  m_fl_variable_alpha = true;
  m_small_threshold = 512;
  */
  m_alpha = 0.6;
  m_amount = 0.1;
  m_radius = 1.0;
  m_smooth_coef = 2000000;
  m_hue_num_of_iter = 3;
  m_diff_num_of_iter = 3;
  m_fl_variable_alpha = false;
  m_small_threshold = 512;
  }
//-----------------------------------------------------------------------------
WeightedEdgeImageProcessor& WeightedEdgeImageProcessor::GetImageProcessor()
  {
  static WeightedEdgeImageProcessor we_ip;
  return we_ip;
  }
//-----------------------------------------------------------------------------
void WeightedEdgeImageProcessor::GetParameters(Params& o_params) const
  {
  o_params["WEdge_Alpha"] = boost::lexical_cast<std::string>(m_alpha);
  o_params["WEdge_Amount"] = boost::lexical_cast<std::string>(m_amount);
  o_params["WEdge_Radius"] = boost::lexical_cast<std::string>(m_radius);
  o_params["WEdge_DiffusionIterations"] = boost::lexical_cast<std::string>(m_diff_num_of_iter);
  o_params["WEdge_HueIterations"] = boost::lexical_cast<std::string>(m_hue_num_of_iter);
  o_params["WEdge_SmallThreshold"] = boost::lexical_cast<std::string>(m_small_threshold);
  }
//-----------------------------------------------------------------------------
void WeightedEdgeImageProcessor::GetDescriptions(Params& o_params) const
  {
  o_params["WeightedEdgeImageProcessor::CorrectHue()"] = "Corrects image colors by selective hue averaging";
  o_params["WeightedEdgeImageProcessor::ApplyDiffusionFlow()"] = "Applies anisotropic diffusion flow (denoising+edge enhancing)";

  o_params["WEdge_Alpha"] = "Coefficient that controls diffusion: "
    "values closer to 0.1 gives smoother result, to 1.0 - sharper. "
    "If you set it closer to 0.1 - also increase number of diffusion iterations";
  o_params["WEdge_Amount"] = "Coeficient that controls amount of diffusion. "
    "If you increase it - also increase number of diffusion iterations";
  o_params["WEdge_Radius"] = "Radius that controls diffusion spreading. "
    "Higher values give more edge-directed result";
  o_params["WEdge_DiffusionIterations"] = "Number of diffusion iterations. "
    "Does not control diffusion strength (see amount for that). You should not "
    "set this value too small, otherwise diffusion process will be unstable";
  o_params["WEdge_HueIterations"] = "Number of iterations on hue color correction "
    "step. Higher values (up to some threshold) usually give less chroma artifacts, "
    "but produce more noise";
  o_params["WEdge_SmallThreshold"] = "If pixel value is smaller "
    "than this threshold value - safer interpolation method will be used for current pixel";
  }
//-----------------------------------------------------------------------------
void WeightedEdgeImageProcessor::SetParameters(Params& i_params)
  {
  if (i_params.exist("WEdge_Alpha"))
    m_alpha = boost::lexical_cast<float>(i_params.pop("WEdge_Alpha"));
  if (i_params.exist("WEdge_Amount"))
    m_amount = boost::lexical_cast<float>(i_params.pop("WEdge_Amount"));
  if (i_params.exist("WEdge_Radius"))
    m_radius = boost::lexical_cast<float>(i_params.pop("WEdge_Radius"));
  if (i_params.exist("WEdge_DiffusionIterations"))
    m_diff_num_of_iter = boost::lexical_cast<ulong>(i_params.pop("WEdge_DiffusionIterations"));
  if (i_params.exist("WEdge_HueIterations"))
    m_hue_num_of_iter = boost::lexical_cast<ulong>(i_params.pop("WEdge_HueIterations"));
  if (i_params.exist("WEdge_SmallThreshold"))
    m_small_threshold = boost::lexical_cast<ulong>(i_params.pop("WEdge_SmallThreshold"));
  }
//-----------------------------------------------------------------------------
void WeightedEdgeImageProcessor::Process(Image& image, const RawImageInfo& image_info, const std::string& method)
  {
  if (method == "CorrectHue()")
    CorrectHue(image, image_info);
  else if (method == "ApplyDiffusionFlow()")
    ApplyDiffusionFlow(image, image_info);
  else throw ExImageProcessorError("UnknMethod", "Attempting to call unexisting processing method WeightedEdgeImageProcessor::" + method);
  }
//-----------------------------------------------------------------------------
float WeightedEdgeImageProcessor::CorrectGreenByChroma(Image& image, unsigned row, unsigned col, uchar ind, float coefs[])
  {
  float green;

  if (image.At(row - 1, col)[ind] < m_small_threshold ||
      image.At(row + 1, col)[ind] < m_small_threshold ||
      image.At(row, col - 1)[ind] < m_small_threshold ||
      image.At(row, col + 1)[ind] < m_small_threshold)
    {
    green = image.At(row, col)[1];
    }
  else
    {
    green = image.At(row, col)[ind] *
	    (coefs[0]*(float)image.At(row - 1, col)[1]/image.At(row - 1, col)[ind]
	     + coefs[1]*(float)image.At(row, col - 1)[1]/image.At(row, col - 1)[ind]
	     + coefs[2]*(float)image.At(row, col + 1)[1]/image.At(row, col + 1)[ind]
	     + coefs[3]*(float)image.At(row + 1, col)[1]/image.At(row + 1, col)[ind]) /
	    (coefs[0] + coefs[1] + coefs[2] + coefs[3]);
    }

  return green;
  }
//-----------------------------------------------------------------------------
float WeightedEdgeImageProcessor::CorrectGreen(Image& image, unsigned row, unsigned col, float coefs[])
  {
  return (CorrectGreenByChroma(image, row, col, 0, coefs)
	  + CorrectGreenByChroma(image, row, col, 2, coefs)) / 2.0;
  }
//-----------------------------------------------------------------------------
float WeightedEdgeImageProcessor::CorrectChroma(Image& image, unsigned row, unsigned col, uchar ind, float coefs[])
  {
  float chroma;

  if (image.At(row - 1, col)[1] < m_small_threshold ||
      image.At(row + 1, col)[1] < m_small_threshold ||
      image.At(row, col - 1)[1] < m_small_threshold ||
      image.At(row, col + 1)[1] < m_small_threshold ||
      image.At(row - 1, col - 1)[1] < m_small_threshold ||
      image.At(row - 1, col + 1)[1] < m_small_threshold ||
      image.At(row + 1, col - 1)[1] < m_small_threshold ||
      image.At(row + 1, col + 1)[1] < m_small_threshold)
    {
    chroma = image.At(row, col)[ind];
    }
  else
    {
    chroma = image.At(row, col)[1] *
	     (coefs[0]*(float)image.At(row - 1, col)[ind]/image.At(row - 1, col)[1]
	      + coefs[1]*(float)image.At(row, col - 1)[ind]/image.At(row, col - 1)[1]
	      + coefs[2]*(float)image.At(row, col + 1)[ind]/image.At(row, col + 1)[1]
	      + coefs[3]*(float)image.At(row + 1, col)[ind]/image.At(row + 1, col)[1]

	      + coefs[4]*(float)image.At(row - 1, col - 1)[ind]/image.At(row - 1, col - 1)[1]
	      + coefs[5]*(float)image.At(row - 1, col + 1)[ind]/image.At(row - 1, col + 1)[1]
	      + coefs[6]*(float)image.At(row + 1, col - 1)[ind]/image.At(row + 1, col - 1)[1]
	      + coefs[7]*(float)image.At(row + 1, col + 1)[ind]/image.At(row + 1, col + 1)[1]) /
	     (coefs[0] + coefs[1] + coefs[2] + coefs[3]
	      + coefs[4] + coefs[5] + coefs[6] + coefs[7]);
    }

  return chroma;
  }
//-----------------------------------------------------------------------------
void WeightedEdgeImageProcessor::CorrectHue(Image& image, const RawImageInfo& image_info)
  {
  mp_image_info = &image_info;

  float coefs[8];

  for (uchar num = 0; num < m_hue_num_of_iter; ++num)
    {
    mg_log << INFO << "Correcting colors (pass " << num+1 << " of " << m_hue_num_of_iter << ")" << endlog;

    // We should skip trim zone, otherwise we will propagate incorrect values
    unsigned trim = std::max<unsigned>(image.GetTrimWidth(), 2);

    image.StartBuffer(trim, 3);

    for (unsigned row = trim; row < image.GetHeight() - trim; ++row)
      {
      for (unsigned col = trim; col < image.GetWidth() - trim; ++col)
	{
	uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);

	coefs[0] = GetEdgeWeight(image, row, col, 2);
	coefs[1] = GetEdgeWeight(image, row, col, 4);
	coefs[2] = GetEdgeWeight(image, row, col, 6),
	coefs[3] = GetEdgeWeight(image, row, col, 8),

	coefs[4] = GetEdgeWeight(image, row, col, 1),
	coefs[5] = GetEdgeWeight(image, row, col, 3),
	coefs[6] = GetEdgeWeight(image, row, col, 7),
	coefs[7] = GetEdgeWeight(image, row, col, 9);

	for (uchar i = 0; i < 3; ++i)
	  if (i != ind)
	    if (i == 1)
	      image.AtBuffer(row, col)[1] =
		image.ClampChromaPixel((int)CorrectGreen(image, row, col, coefs), row, col);
	    else
	      image.AtBuffer(row, col)[i] =
		image.ClampChromaPixel((int)CorrectChroma(image, row, col, i, coefs), row, col);
	}
      image.FlushBufferRow();
      }

    image.StopBuffer();
    }
  }
//-----------------------------------------------------------------------------
void WeightedEdgeImageProcessor::ComputeG(const Image& img_src, unsigned row, unsigned col,
					  MatrixG& g)
  {
  float r_x = ((int)img_src.At(row, col - 1)[0] - (int)img_src.At(row, col + 1)[0]) / 2.0;
  float g_x = ((int)img_src.At(row, col - 1)[1] - (int)img_src.At(row, col + 1)[1]) / 2.0;
  float b_x = ((int)img_src.At(row, col - 1)[2] - (int)img_src.At(row, col + 1)[2]) / 2.0;

  float r_y = ((int)img_src.At(row - 1, col)[0] - (int)img_src.At(row + 1, col)[0]) / 2.0;
  float g_y = ((int)img_src.At(row - 1, col)[1] - (int)img_src.At(row + 1, col)[1]) / 2.0;
  float b_y = ((int)img_src.At(row - 1, col)[2] - (int)img_src.At(row + 1, col)[2]) / 2.0;

  g.AtBuffer(row, col)[0] = 1 + r_x*r_x + g_x*g_x + b_x*b_x;
  g.AtBuffer(row, col)[1] = r_x*r_y + g_x*g_y + b_x*b_y;
  g.AtBuffer(row, col)[2] = 1 + r_y*r_y + g_y*g_y + b_y*b_y;
  }
//-----------------------------------------------------------------------------
void WeightedEdgeImageProcessor::CreateGaussian(Gaussian& gs, float r)
  {
  int center = gs.GetWidth() / 2;
  float sum = 0;

  for (int row = 0; row < gs.GetWidth(); ++row)
    for (int col = 0; col < gs.GetWidth(); ++col)
      {
      float val = exp(-(sqr(row - center) + sqr(col - center))/sqr(r));
      gs.At(row, col) = val;
      sum += val;
      }

  for (int row = 0; row < gs.GetWidth(); ++row)
    for (int col = 0; col < gs.GetWidth(); ++col)
      gs.At(row, col) /= sum;
  }
//-----------------------------------------------------------------------------
void WeightedEdgeImageProcessor::DiffuseG(const Image& img,
					  const MatrixG& g_src,
					  MatrixG& g_dest,
					  unsigned row,
					  unsigned col,
					  const Gaussian& gs)
  {
  unsigned w2 = gs.GetWidth()/2, rw = row - w2, cw = col - w2;
  float s[3] = {0, 0, 0};

  for (int r = 0; r < gs.GetWidth(); ++r)
    for (int c = 0; c < gs.GetWidth(); ++c)
      {
      const PixelTmpl<float, 3>& pix = g_src.AtBuffer(rw + r, cw + c);
      float v = gs.At(r, c);
      s[0] += v*pix[0];
      s[1] += v*pix[1];
      s[2] += v*pix[2];
      }

  PixelTmpl<float, 3>& pix = g_dest.AtBuffer(row, col);

  pix[0] = s[0];
  pix[1] = s[1];
  pix[2] = s[2];
  }
//-----------------------------------------------------------------------------
void WeightedEdgeImageProcessor::ApplyStep(Image& img,
					   unsigned row,
					   unsigned col,
					   const MatrixG& g,
					   float step)
  {
  float a = g.AtBuffer(row, col)[0];
  float b = g.AtBuffer(row, col)[1];
  float c = g.AtBuffer(row, col)[2];

  float u[4], uT[4];
  float l1, l2;

  if (fabs(b) < 0.05) // Assume b == 0
    {
    u[0] = uT[0] = 1.0;
    u[1] = uT[1] = 0.0;
    u[2] = uT[2] = 0.0;
    u[3] = uT[3] = 1.0;

    l1 = a;
    l2 = c;
    }
  else
    {
    float amc = a - c, apc = a + c, d = sqrt(sqr(amc) + 4*sqr(b));

    // Let's decompose matrix:
    l1 = 0.5*(apc - d);
    l2 = 0.5*(apc + d);

    u[0] = (amc - d)/(2*b);
    u[1] = 1;
    u[2] = (amc + d)/(2*b);
    u[3] = 1;

    uT[0] = -b/d;
    uT[1] = b/d;
    uT[2] = (amc + d)/(2*d);
    uT[3] = - (amc - d)/(2*d);
    }

  float alpha, beta, new_step;

  if (!m_fl_variable_alpha || m_alpha < 0.1)
    {
    alpha = m_alpha;
    beta = -m_alpha;
    new_step = step;
    }
  else
    {
    alpha = 0.1 + (m_alpha - 0.1)*std::max(l1, l2)/m_smooth_coef;
    beta = -alpha;
    new_step = step;

    if (alpha > m_alpha)
      {
      alpha = m_alpha;
      beta = -m_alpha;
      }
    }

  // Let's change eigenvalues:
  if (l1 > l2)
    {
    l2 = 1/alpha; l1 = beta;
    }
  else
    {
    l1 = 1/alpha; l2 = beta;
    }

  // Let's compose the matrix back:

  float m[4] = {u[0]*l1, u[1]*l1, u[2]*l2, u[3]*l2};

  float mn[4] = {m[0]*uT[0] + m[2]*uT[1],
		  m[1]*uT[0] + m[3]*uT[1],
		  m[0]*uT[2] + m[2]*uT[3],
		  m[1]*uT[2] + m[3]*uT[3]};

  // Let's pass channels through equation:
  for (uchar ind = 0; ind < 3; ++ind)
    {
    float dxx = (int)img.At(row, col - 1)[ind] - 2*(int)img.At(row, col)[ind] + (int)img.At(row, col + 1)[ind],
	  dyy = (int)img.At(row - 1, col)[ind] - 2*(int)img.At(row, col)[ind] + (int)img.At(row + 1, col)[ind],
	  dxy = (-(int)img.At(row + 1, col - 1)[ind] + (int)img.At(row + 1, col + 1)[ind]
		 + (int)img.At(row - 1, col - 1)[ind] - (int)img.At(row - 1, col + 1)[ind]) / 4;

    img.AtBuffer(row, col)[ind] =
      img.ClampChromaPixel((int)((int)img.At(row, col)[ind] +
			   new_step*(mn[0] * dxx + mn[1] * dxy + mn[2] * dxy + mn[3] * dyy)),
			   row, col);
    }
  }
//-----------------------------------------------------------------------------
void WeightedEdgeImageProcessor::ApplyDiffusionFlow(Image& image, const RawImageInfo& image_info)
  {
  mp_image_info = &image_info;

  // Looks strange, no? ;-)
  // But don't be afraid - only their buffers will be used
  MatrixG g(image.GetWidth(), 0);
  MatrixG g_diffused(image.GetWidth(), 0);

  unsigned gs_width;
  if (m_radius <= 2)
    gs_width = 5;
  else
    gs_width = (unsigned)(m_radius * 3 - 1);

  Gaussian gaussian(gs_width, gs_width);
  CreateGaussian(gaussian, m_radius);

  float step = m_amount/m_diff_num_of_iter;

  for (uchar num = 0; num < m_diff_num_of_iter; ++num)
    {
    mg_log << INFO << "Applying diffusion flow (pass " << num+1 << " of " << m_diff_num_of_iter << ")" << endlog;

    unsigned trim = 2;

    // We should accumulate G first on `gs_width' number of rows
    g.StartBuffer(trim, gs_width + 1, true);
    for (unsigned row = trim; row < trim + gs_width; ++row)
      {
      for (unsigned col = trim; col < image.GetWidth() - trim; ++col)
	ComputeG(image, row, col, g);

      g.FlushBufferRow();
      }

    g_diffused.StartBuffer(trim + gs_width/2, 1, true);
    image.StartBuffer(trim + gs_width/2, 3);

    for (unsigned row = trim + gs_width/2; row < image.GetHeight() - trim - gs_width/2; ++row)
      {
      for (unsigned col = trim + gs_width/2; col < image.GetWidth() - trim - gs_width/2; ++col)
	{
	DiffuseG(image, g, g_diffused, row, col, gaussian);
	ApplyStep(image, row, col, g_diffused, step);
	ComputeG(image, row + gs_width/2 + 1, col, g);
	}

      g.FlushBufferRow();
      g_diffused.FlushBufferRow();
      image.FlushBufferRow();
      }

    g.StopBuffer();
    g_diffused.StopBuffer();
    image.StopBuffer();
    }
  }
//-----------------------------------------------------------------------------
