/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

/*****************************************************************************
 * WeightedEdgeInterpolator - interpolation using weighted sums,
 * based on description in ``Demosaicing: Image reconstruction from color
 * CCD samples'' by Ron Kimmel.
******************************************************************************/

#include "stdafx.h"

#include "WeightedEdgeInterpolator.h"
#include "Image.h"
#include "RawImageInfo.h"
//-----------------------------------------------------------------------------
WeightedEdgeInterpolator::WeightedEdgeInterpolator(): IParametrizable("WeightedEdgeInterpolator")
  {
  m_small_threshold = 512;
  }
//-----------------------------------------------------------------------------
WeightedEdgeInterpolator& WeightedEdgeInterpolator::GetInterpolator()
  {
  static WeightedEdgeInterpolator we_int;
  return we_int;
  }
//-----------------------------------------------------------------------------
void WeightedEdgeInterpolator::GetParameters(Params& o_params) const
  {
  o_params["WEdge_SmallThreshold"] = boost::lexical_cast<std::string>(m_small_threshold);
  }
//-----------------------------------------------------------------------------
void WeightedEdgeInterpolator::GetDescriptions(Params& o_params) const
  {
  o_params["WEdge_SmallThreshold"] = "If pixel value is smaller "
    "than this threshold value - safer interpolation method will be used for current pixel";
  }
//-----------------------------------------------------------------------------
void WeightedEdgeInterpolator::SetParameters(Params& i_params)
  {
  if (i_params.exist("WEdge_SmallThreshold"))
    m_small_threshold = boost::lexical_cast<ulong>(i_params.pop("WEdge_SmallThreshold"));
  }
//-----------------------------------------------------------------------------
void WeightedEdgeInterpolator::InterpolateGreenPixel(Image& image, unsigned row, unsigned col)
  {
  assert (mp_image_info -> GetCurrentColorIndex(row, col) != 1);

  float coefs[4] = {GetEdgeWeight(image, row, col, 2),
		     GetEdgeWeight(image, row, col, 4),
		     GetEdgeWeight(image, row, col, 6),
		     GetEdgeWeight(image, row, col, 8)};

  int green = (int)((coefs[0]*image.At(row - 1, col)[1]
		    + coefs[1]*image.At(row, col - 1)[1]
		    + coefs[2]*image.At(row, col + 1)[1]
		    + coefs[3]*image.At(row + 1, col)[1]) /
		    (coefs[0] + coefs[1] + coefs[2] + coefs[3]));

  image.At(row, col)[1] = image.ClampChromaPixel(green, row, col);
  }
//-----------------------------------------------------------------------------
void WeightedEdgeInterpolator::InterpolateChromaPixelPass1(Image& image, unsigned row, unsigned col)
  {
  uchar ind = mp_image_info -> GetCurrentColorIndex(row - 1, col - 1);

  float coefs[4] = {GetEdgeWeight(image, row, col, 1),
		     GetEdgeWeight(image, row, col, 3),
		     GetEdgeWeight(image, row, col, 7),
		     GetEdgeWeight(image, row, col, 9)};

  int chroma;

  if (image.At(row - 1, col - 1)[1] < m_small_threshold ||
      image.At(row - 1, col + 1)[1] < m_small_threshold ||
      image.At(row + 1, col - 1)[1] < m_small_threshold ||
      image.At(row + 1, col + 1)[1] < m_small_threshold)
    {
    chroma = (int)((coefs[0]*(float)image.At(row - 1, col - 1)[ind]
		    + coefs[1]*(float)image.At(row - 1, col + 1)[ind]
		    + coefs[2]*(float)image.At(row + 1, col - 1)[ind]
		    + coefs[3]*(float)image.At(row + 1, col + 1)[ind]) /
		   (coefs[0] + coefs[1] + coefs[2] + coefs[3]));
    }
  else
    {
    chroma = (int)(image.At(row, col)[1] *
		   (coefs[0]*(float)image.At(row - 1, col - 1)[ind]/image.At(row - 1, col - 1)[1]
		    + coefs[1]*(float)image.At(row - 1, col + 1)[ind]/image.At(row - 1, col + 1)[1]
		    + coefs[2]*(float)image.At(row + 1, col - 1)[ind]/image.At(row + 1, col - 1)[1]
		    + coefs[3]*(float)image.At(row + 1, col + 1)[ind]/image.At(row + 1, col + 1)[1]) /
		   (coefs[0] + coefs[1] + coefs[2] + coefs[3]));
    }

  image.At(row, col)[ind] = image.ClampChromaPixel(chroma, row, col);
  }
//-----------------------------------------------------------------------------
void WeightedEdgeInterpolator::InterpolateChromaPixelPass2(Image& image, unsigned row, unsigned col, uchar ind)
  {
  float coefs[4] = {GetEdgeWeight(image, row, col, 2),
		     GetEdgeWeight(image, row, col, 4),
		     GetEdgeWeight(image, row, col, 6),
		     GetEdgeWeight(image, row, col, 8)};

  int chroma;

  if (image.At(row - 1, col)[1] < m_small_threshold ||
      image.At(row + 1, col)[1] < m_small_threshold ||
      image.At(row, col - 1)[1] < m_small_threshold ||
      image.At(row, col + 1)[1] < m_small_threshold)
    {
    chroma = (int)((coefs[0]*(float)image.At(row - 1, col)[ind]
		    + coefs[1]*(float)image.At(row, col - 1)[ind]
		    + coefs[2]*(float)image.At(row, col + 1)[ind]
		    + coefs[3]*(float)image.At(row + 1, col)[ind]) /
		   (coefs[0] + coefs[1] + coefs[2] + coefs[3]));
    }
  else
    {
    chroma = (int)(image.At(row, col)[1] *
		   (coefs[0]*(float)image.At(row - 1, col)[ind]/image.At(row - 1, col)[1]
		    + coefs[1]*(float)image.At(row, col - 1)[ind]/image.At(row, col - 1)[1]
		    + coefs[2]*(float)image.At(row, col + 1)[ind]/image.At(row, col + 1)[1]
		    + coefs[3]*(float)image.At(row + 1, col)[ind]/image.At(row + 1, col)[1]) /
		   (coefs[0] + coefs[1] + coefs[2] + coefs[3]));
    }

  image.At(row, col)[ind] = image.ClampChromaPixel(chroma, row, col);
  }
//-----------------------------------------------------------------------------
bool WeightedEdgeInterpolator::Interpolate(Image& image, const RawImageInfo& image_info)
  {
  mg_log << INFO << "Running Weighted Edge interpolation" << endlog;
  mp_image_info = &image_info;

  mg_log << INFO << "Interpolating green" << endlog;

  for (unsigned row = 2; row < image.GetHeight() - 2; ++row)
    for (unsigned col = 2; col < image.GetWidth() - 2; ++col)
      if ( mp_image_info -> GetCurrentColorIndex(row, col) != 1)
	InterpolateGreenPixel(image, row, col);

  image.SetState(Image::FULL_GREEN_COLORS);

  mg_log << INFO << "Interpolating chroma (1st pass)" << endlog;

  for (unsigned row = 2; row < image.GetHeight() - 2; ++row)
    for (unsigned col = 2; col < image.GetWidth() - 2; ++col)
      {
      uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);

      if (ind != 1)
	InterpolateChromaPixelPass1(image, row, col);
      }

  mg_log << INFO << "Interpolating chroma (2nd pass)" << endlog;

  for (unsigned row = 2; row < image.GetHeight() - 2; ++row)
    for (unsigned col = 2; col < image.GetWidth() - 2; ++col)
      {
      uchar ind = mp_image_info -> GetCurrentColorIndex(row, col);

      if (ind != 1)
	continue;

      InterpolateChromaPixelPass2(image, row, col, 0);
      InterpolateChromaPixelPass2(image, row, col, 2);
      }

  image.SetState(Image::FULL_COLORS);

  mg_log << TRACE << "Finished" << endlog;

  return true;
  }
//-----------------------------------------------------------------------------
