/*****************************************************************************
 *
 * LibUnrawel, Copyright (C) 2002-2003 by Alexander Tsvyashchenko
 * contact address: <spam@ndl.kiev.ua>
 *
 * short license info is below, but it concerns only program itself -
 * you should see copyrights.txt for algorithms licenses info!
 *
 *****************************************************************************

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

******************************************************************************/

#ifndef LOGSTREAM_H
#define LOGSTREAM_H

#include <sstream>
//--------------------------------------------------------------------
// Example of usage:
//
// mg_log.SetOutput(std::clog);
// mg_log.SetLevel(INFO);
// mg_log.SetAddTime(true);
//
// mg_log.AddOutput(std::cout);
// mg_log.SetLevel(WARN);
// mg_log.SetAddTime(false);
// ...
// mg_log << TRACE << "text1" << "text2" << endlog;
// ...
// if (critical_error)
//   {
//   mg_log.PrintAllRecentLogMessages();
//   exit(1);
//   }
//--------------------------------------------------------------------
class LogErrorLevel
  {
  public:
    explicit LogErrorLevel(int level) { m_level = level; }

    int GetErrorLevel() { return m_level; }

  private:
    int m_level;
  };
//--------------------------------------------------------------------
// I think it's safe to put these names in global scope ...
extern LogErrorLevel CRIT, ERR, WARN, INFO, TRACE, DEBUG;
//--------------------------------------------------------------------
class logstream: public std::stringstream
  {
  friend std::ostream& endlog(std::ostream& os);

  public:
    logstream();

    ~logstream();

    // Per-output setting
    void SetLevel(LogErrorLevel i_lev);

    // Set i_fl_is_responsible if logstream should delete the stream in destructor
    void SetOutput(std::ostream& i_out, bool i_fl_is_responsible = false);

    // Set i_fl_is_responsible if logstream should delete the stream in destructor
    void AddOutput(std::ostream& i_out, bool i_fl_is_responsible = false);

    // Per-output setting
    void SetAddTime(bool i_fl_add_time);

    // Per-output setting
    void SetAddLevel(bool i_fl_add_level);

    // Global setting
    void SetMaxHistoryLines(unsigned long i_lines);

    // Prints all recent log messages of all levels
    void PrintAllRecentLogMessages();

    std::ostream& operator << (LogErrorLevel i_lev);

  private:
    struct LogInfo
      {
      LogErrorLevel m_desired_level;
      bool m_fl_add_time;
      bool m_fl_add_level;
      mutable bool m_fl_is_responsible;
      std::ostream* mp_out;

      LogInfo(std::ostream* ip_out, bool i_fl_is_responsible): m_desired_level(INFO)
        {
        m_fl_add_time = true;
        m_fl_add_level = true;
        mp_out = ip_out;
        m_fl_is_responsible = i_fl_is_responsible;
        }

      LogInfo(const LogInfo& i_log_info): m_desired_level(INFO)
        {
        *this = i_log_info;
        }

      LogInfo& operator =(const LogInfo& i_log_info)
        {
        m_fl_add_time = i_log_info.m_fl_add_time;
        m_fl_add_level = i_log_info.m_fl_add_level;
        mp_out = i_log_info.mp_out;
        m_fl_is_responsible = i_log_info.m_fl_is_responsible;
        i_log_info.m_fl_is_responsible = false;

        return *this;
        }

      ~LogInfo()
        {
        if (m_fl_is_responsible)
          delete mp_out;
        }
      };

    // By design it never can be empty
    std::list<LogInfo> m_outs;
    std::list<std::string> m_history;
    unsigned long m_max_history_lines;
    LogErrorLevel m_curr_level;
    std::string m_curr_time;
    static std::string mg_level_str[6];

  private:
    std::string FormatMessage(const std::string& i_str, bool fl_add_time, bool fl_add_level);
  };
//--------------------------------------------------------------------
inline void logstream::SetLevel(LogErrorLevel i_lev)
  {
  m_outs.back().m_desired_level = i_lev;
  }
//--------------------------------------------------------------------
inline void logstream::SetOutput(std::ostream& i_out, bool i_fl_is_responsible)
  {
  m_outs.clear();
  AddOutput(i_out, i_fl_is_responsible);
  }
//--------------------------------------------------------------------
inline void logstream::AddOutput(std::ostream& i_out, bool i_fl_is_responsible)
  {
  m_outs.push_back(LogInfo(&i_out, i_fl_is_responsible));
  }
//--------------------------------------------------------------------
inline void logstream::SetAddTime(bool i_fl_add_time)
  {
  m_outs.back().m_fl_add_time = i_fl_add_time;
  }
//--------------------------------------------------------------------
inline void logstream::SetAddLevel(bool i_fl_add_level)
  {
  m_outs.back().m_fl_add_level = i_fl_add_level;
  }
//--------------------------------------------------------------------
inline void logstream::SetMaxHistoryLines(unsigned long i_lines)
  {
  m_max_history_lines = i_lines;
  }
//--------------------------------------------------------------------
extern logstream mg_log;
//--------------------------------------------------------------------

#endif // LOGSTREAM_H
